'''Autogenerated by xml_generate script, do not edit!'''
from OpenGL import platform as _p, arrays
# Code generation uses this
from OpenGL.raw.GL import _types as _cs
# End users want this...
from OpenGL.raw.GL._types import *
from OpenGL.raw.GL import _errors
from OpenGL.constant import Constant as _C

import ctypes
_EXTENSION_NAME = 'GL_NV_register_combiners2'
def _f( function ):
    return _p.createFunction( function,_p.PLATFORM.GL,'GL_NV_register_combiners2',error_checker=_errors._error_checker)
GL_PER_STAGE_CONSTANTS_NV=_C('GL_PER_STAGE_CONSTANTS_NV',0x8535)
@_f
@_p.types(None,_cs.GLenum,_cs.GLenum,arrays.GLfloatArray)
def glCombinerStageParameterfvNV(stage,pname,params):pass
# Calculate length of params from pname:CombinerParameterNV
@_f
@_p.types(None,_cs.GLenum,_cs.GLenum,arrays.GLfloatArray)
def glGetCombinerStageParameterfvNV(stage,pname,params):pass
# Calculate length of params from pname:CombinerParameterNV



from fsgs.platform import PlatformHandler
from fsgs.mednafen.nintendo import NintendoRunner
from .loader import SimpleLoader


class NintendoPlatformHandler(PlatformHandler):

    PLATFORM_NAME = "Nintendo"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return NintendoLoader(fsgs)

    def get_runner(self, fsgs):
        return NintendoRunner(fsgs)


class NintendoLoader(SimpleLoader):
    pass



from fsui import Application as BaseApplication
from fsgs.application import ApplicationMixin


class Application(ApplicationMixin, BaseApplication):

    instance = None
    name = None

    @classmethod
    def init(cls, name):
        cls.name = name

    def __init__(self):
        BaseApplication.__init__(self, Application.name)



'''OpenGL extension VERSION.GL_1_2

This module customises the behaviour of the 
OpenGL.raw.GL.VERSION.GL_1_2 to provide a more 
Python-friendly API

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/VERSION/GL_1_2.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
from OpenGL.GL import glget
import ctypes
from OpenGL.raw.GL import _types
from OpenGL.raw.GL.VERSION.GL_1_2 import *
from OpenGL.raw.GL.VERSION.GL_1_2 import _EXTENSION_NAME

def glInitGl12VERSION():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

### END AUTOGENERATED SECTION
from OpenGL.GL.ARB.imaging import *
from OpenGL.GL.VERSION.GL_1_2_images import *

GL_POINT_SIZE_GRANULARITY = GL_SMOOTH_POINT_SIZE_GRANULARITY # alias
GL_POINT_SIZE_RANGE = GL_SMOOTH_POINT_SIZE_RANGE # alias
GL_LINE_WIDTH_GRANULARITY = GL_SMOOTH_LINE_WIDTH_GRANULARITY # alias
GL_LINE_WIDTH_RANGE = GL_SMOOTH_LINE_WIDTH_RANGE # alias

glget.addGLGetTexParameterConstant(GL_TEXTURE_MIN_LOD,(1,))
glget.addGLGetTexParameterConstant(GL_TEXTURE_MAX_LOD,(1,))
glget.addGLGetTexParameterConstant(GL_TEXTURE_BASE_LEVEL,(1,))
glget.addGLGetTexParameterConstant(GL_TEXTURE_MAX_LEVEL,(1,))
glget.addGLGetTexParameterConstant(GL_TEXTURE_WRAP_R,(1,))

glDrawRangeElements = wrapper.wrapper( glDrawRangeElements ).setPyConverter(
    'indices', arrays.AsArrayOfType( 'indices', 'type' ),
).setReturnValues(
    wrapper.returnPyArgument( 'indices' )
)





from fsui.qt import QMenu


class Menu:

    def __init__(self):
        self.qmenu = QMenu()
        # self._menu = wx.Menu()
        # self._ids = []
        # #self._functions = []
        pass

    def add_item(self, text, function, item_id=-1):
        text = text.replace("&", "&&")

        action = self.qmenu.addAction(text)
        action.triggered.connect(function)

        # if item_id == -1:
        #     item_id = wx.NewId()
        # self._ids.append(item_id)
        # #self._functions.append(function)
        # self._menu.Append(item_id, text)
        # def function_wrapper(event):
        #     # the reason we use CallAfter here is so control will return
        #     # to the application after having popped up a menu, so the
        #     # button opening a menu can draw its correct state when the
        #     # menu is closed, before for example a modal dialog is opened.
        #     wx.CallAfter(function)
        # self._menu.Bind(wx.EVT_MENU, function_wrapper, id=item_id)

    def add_about_item(self, text, function):
        self.add_item(text, function)

    def add_preferences_item(self, text, function):
        self.add_item(text, function)

    def add_separator(self):
        self.qmenu.addSeparator()



import os
import io
from configparser import ConfigParser
from fsbc.system import windows, linux, macosx
from fsbc.util import memoize
from fsgs.context import fsgs
from fsbc.Resources import Resources


class InputDeviceNotFoundException(Exception):
    pass


class MissingPlatformSupportException(Exception):
    pass


class InputDeviceInfo(object):
    id = ""
    generic_id = ""
    sdl_name = ""
    name = ""
    buttons = 0
    axes = 0
    hats = 0
    balls = 0
    index = 0

    def get_type(self):
        if self.sdl_name:
            return 'TYPE_JOYSTICK'
        else:
            return 'TYPE_KEYBOARD'


class InputDevice(object):

    MissingPlatformSupportException = MissingPlatformSupportException 

    def __init__(self, platform, name, sclist, sdl_name="", sdl_joystick_id=-1,
                 version=1, buttons=0, axes=0, hats=0, balls=0):
        """
        sclist -- system controller list
        """
        self.type = ""
        self.config = None
        self.config_inv = None
        self.id = name
        # self._name = name
        if "#" in self.id:
            self.name, dummy = self.id.rsplit('#', 1)
        else:
            self.name = self.id
        self.name = self.name.strip()
        self.decorate_name_with_number()

        # print("--------->", self.id)
        # print("------------>", self.name)
        # print("********** InputDevice Constructor", self.name, "ID is",
        #         self.id)

        self.platform = platform
        self.sdl_name = sdl_name
        if not self.sdl_name:
            self.sdl_name = self.id.rsplit("#", 1)[0]
        self.sdl_joystick_id = sdl_joystick_id
        self.index = -1
        self.version = version
        self.buttons = buttons
        self.axes = axes
        self.hats = hats
        self.balls = balls
        for sc in sclist:
            # print("---", name, "---", sc.id, sc.index)
            if name == sc.id:
                self.index = sc.index
                break

        if version == 1:
            method_name = "configure_for_" + platform.lower().replace(
                " ", "_").replace("-", "_")
            if hasattr(self, method_name):
                getattr(self, method_name)(name, sclist=sclist)
            else:
                print("{0} does not support platform {1}".format(
                    name, platform))
                raise MissingPlatformSupportException(
                    "No input device support for platform")
        elif version == 2:
            old_name = self.name
            self.configure_version_2()
            if self.name != old_name:
                # print("****** name is now", self.name)
                print(self.id, "=>", repr(self.name))

    def decorate_name_with_number(self):
        # print("decorate_name_with_number, ID is", self.id)
        if "#" in self.id:
            dummy, number_str = self.id.rsplit('#', 1)
        else:
            number_str = "1"
        if number_str != '1':
            self.name = "{0} #{1}".format(self.name, number_str)

    def is_joystick(self):
        return self.get_type() == 'TYPE_JOYSTICK'

    def is_keyboard(self):
        return self.get_type() == 'TYPE_KEYBOARD'

    def get_type(self):
        if self.type.startswith('joystick'):
            return 'TYPE_JOYSTICK'
        elif self.type.startswith('keyboard'):
            return 'TYPE_KEYBOARD'
        else:
            raise Exception("unknown input device type " + repr(self.type))

    @classmethod
    def get_builtin_config_for_device_guid(cls, guid):
        return Resources("fsgs").stream(
            "res/input/" + guid + ".fs-uae-controller")

    # @classmethod
    # def get_builtin_config_file_for_device_guid(cls, guid):
    #     try:
    #         path = Resources("fsgs").path(
    #             "res/input/" + guid + ".fs-uae-controller")
    #     except LookupError:
    #         return None
    #     return path

    @staticmethod
    @memoize
    def get_config_files():
        print("get_config_files")
        configs = {}
        input_stream = Resources("fsgs").stream("res/input/manifest.txt")
        print("opened input manifest")
        for line in input_stream.read().split(b"\n"):
            line = line.decode("UTF-8")
            path = line.strip()
            if not path:
                continue
            _, ext = os.path.splitext(path)
            if ext in [".ini", ".conf"]:
                parts = path.split("/")
                file_name = parts[-1]
                name, _ = os.path.splitext(file_name)
                if len(parts) > 1:
                    configs[parts[-2] + '_' + name] = "fsgs:res/input/" + path
                print(" -", path)
                configs[name] = "fsgs:res/input/" + path

        # FIXME: fix dependency
        controllers_dir = fsgs.amiga.get_controllers_dir()
        print("read configs from controllers_dir at", controllers_dir)
        if os.path.exists(controllers_dir):
            for file_name in os.listdir(controllers_dir):
                if file_name.endswith(".conf"):
                    name, ext = os.path.splitext(file_name)
                    path = os.path.join(controllers_dir, file_name)
                    print(" -", path)
                    configs[name] = path
        # print("input config files:", configs)
        return configs

    def read_config(self, config_name, config, platform, multiple):
        print("read_config", config_name)
        configs = self.get_config_files()
        print(config_name in configs)
        try:
            path = configs[config_name]
        except KeyError:
            # traceback.print_stack()
            print("no config file found for", repr(self.sdl_name),
                  "=", config_name)
            # raise InputDeviceNotFoundException(
            #         "no config found for " + repr(self.sdl_name))
            if platform:
                raise MissingPlatformSupportException(
                    "no config found for " + repr(self.sdl_name))
            else:
                return
        cp = ConfigParser()
        print(path, os.path.exists(path))
        if path.startswith("fsgs:"):
            print("reading config from stream", path)
            input_stream = Resources("fsgs").stream(path.split(":", 1)[1])
            input_stream = io.TextIOWrapper(input_stream, "UTF-8")
            cp.read_file(input_stream)
        else:
            cp.read(path)
        if cp.has_option('device', 'type'):
            self.type = cp.get('device', 'type')
        if cp.has_option('device', 'name'):
            # print("HAD NAME", self.name, self.sdl_name)
            self.name = cp.get('device', 'name')
            self.decorate_name_with_number()
            # print("HAS NAME", self.name, self.sdl_name)
            # try:
            #     dummy, num = self.name.rsplit('#', 1)
            # except ValueError:
            #     self.name = name
            # else:
            #     self.name = "{0} #{1}".format(name, num)
        if cp.has_section(platform):
            section = platform
        elif cp.has_section('default'):
            # print("has default section")
            section = 'default'
        else:
            if platform:
                raise MissingPlatformSupportException(
                    "no config found for platform " + repr(platform))
            else:
                return
        # config = {}
        # if section == 'gamepad':
        #
            # for option in cp.options('gamepad'):
            #     value = cp.get('gamepad', option)
            #     print("gamepad", option, value)

        if cp.has_option(section, 'include'):
            include_config = cp.get(section, 'include')
            include_config = include_config.replace('/', '_')
            self.read_config(include_config, config, platform, multiple)

        # iconfig = {}
        # for key, value in config.items():
        #     iconfig[value] = key

        # for key in cp.options(section):
        #     value = cp.get(section, key)
        for key, value in cp.items(section):
            value = value.strip()
            if value.startswith('('):
                if not multiple:
                    continue
                assert value.endswith(')')
                value = value[1:-1]
            # print(key, "===>", value)
            # print("key, value is", key, value)
            # if value in iconfig:
            try:
                # config[key] = iconfig[value]
                # config[key] = config[value]
                # del config[value]
                # config[config[value]] = value
                
                # if key in iconfig:
                config[key] = config[value]
                # del iconfig[key]
                # iconfig[config[value]] = key
                # if not config[value] in iconfig:
                #     iconfig[config[value]] = key
                del config[value]
            except KeyError:
                config[key] = value
                # if not value in iconfig:
                #     iconfig[value] = key
                # config_order.append(key)

            # config[option] = cp.get(section, option)

        # if cp.has_section('gamepad'):
        #     for key, value in list(config.items()):
        #         if cp.has_option('gamepad', value):
        #             config[key] = cp.get('gamepad', value)

    def configure_version_2(self):
        self.config = self.configure(self.platform)

    def configure(self, platform, multiple=True):
        print("CONFIGURE", platform, self.name, self.sdl_name)
        # print("InputDevice.configure")
        
        name_lower = self.sdl_name.lower()
        name = ""
        for c in name_lower:
            if c in "abcdefghijklmnopqrstuvwxyz0123456789":
                name = name + c
            else:
                if not name.endswith("_"):
                    name += "_"
        name = name.strip("_")
        if windows:
            host_platform = "windows"
        elif macosx:
            host_platform = "macosx"
        elif linux:
            host_platform = "linux"
        else:
            host_platform = "other"
        config_name = "{0}_{1}_{2}_{3}_{4}_{5}".format(
            name, self.buttons, self.axes, self.hats, self.balls,
            host_platform)

        config = {}
        try:
            self.read_config(config_name, config, platform, multiple)
        except Exception:
            pass
        else:
            return config

        config_name = ""
        for c in self.sdl_name.lower():
            if c in 'abcdefghijklmnopqrstuvwxyz0123456789':
                config_name += c
            elif len(config_name) == 0 or config_name[-1] != '_':
                config_name += '_'
        if config_name.endswith('_usb'):
            config_name = config_name[:-4]
        while config_name.endswith('_'):
            config_name = config_name[:-1]
        # print("config_name =", config_name, "sdl_name", repr(self.sdl_name))
        self.read_config(config_name, config, platform, multiple)
        return config

    @staticmethod
    def supported_keyboard_input_devices():
        return "", 0

    def get_name(self):
        return self.name

    def get_sdl_joystick_id(self):
        return self.sdl_joystick_id

    def get_config(self):
        if self.config is None:
            raise Exception("InputDevice is not configured")
        return self.config

    def get_config_inverted(self):
        if self.config_inv is not None:
            return self.config_inv
        if self.config is None:
            raise Exception("InputDevice is not configured")
        self.config_inv = {}
        for k, v in self.config.items():
            self.config_inv[v] = k
        return self.config_inv



from ..common.layout import HorizontalLayout, VerticalLayout
from ..common.i18n import gettext
from .Button import Button
from .Panel import Panel


class DialogButtons(Panel):

    def __init__(self, parent):
        super().__init__(parent)
        self.layout = HorizontalLayout(padding=20)
        self.layout.padding_top = 0
        self.layout.add_spacer(0, expand=True)

        self.button_layout = HorizontalLayout()
        self.layout.add(self.button_layout, fill=True)

    @classmethod
    def create_with_layout(cls, parent):
        parent.layout = VerticalLayout()
        buttons = DialogButtons(parent)
        layout = VerticalLayout(padding=20)
        parent.layout.add(layout, expand=True, fill=True)
        parent.layout.add(buttons, fill=True)
        return buttons, layout

    def add_button(self, button):
        self.button_layout.add(button, margin_left=10, fill=True)
        return button

    def on_close(self):
        self.get_parent().close()

    def create_close_button(self):
        button = Button(self, gettext("Close"))
        button.activated.connect(self.on_close)
        self.layout.add(button, margin_left=10, fill=True)
        return button



from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from fsgs.mednafen.mednafen import MednafenRunner


class SuperNintendoRunner(MednafenRunner):

    CONTROLLER = {
        "type": "gamepad",
        "description": "Gamepad",
        "mapping_name": "supernintendo",
    }

    PORTS = [
        {
            "description": "1st Controller",
            "types": [CONTROLLER]
        }, {
            "description": "2nd Controller",
            "types": [CONTROLLER]
        },
    ]

    def __init__(self, fsgs):
        super().__init__(fsgs)

    # def mednafen_aspect_ratio(self):
    #     return 4.0 / 3.0

    def mednafen_input_mapping(self, port):
        if port == 0:
            return {
                "A": "snes.input.port1.gamepad.a",
                "B": "snes.input.port1.gamepad.b",
                "X": "snes.input.port1.gamepad.x",
                "Y": "snes.input.port1.gamepad.y",
                "L": "snes.input.port1.gamepad.l",
                "R": "snes.input.port1.gamepad.r",
                "UP": "snes.input.port1.gamepad.up",
                "DOWN": "snes.input.port1.gamepad.down",
                "LEFT": "snes.input.port1.gamepad.left",
                "RIGHT": "snes.input.port1.gamepad.right",
                "SELECT": "snes.input.port1.gamepad.select",
                "START": "snes.input.port1.gamepad.start",
            }
        elif port == 1:
            return {
                "A": "snes.input.port2.gamepad.a",
                "B": "snes.input.port2.gamepad.b",
                "X": "snes.input.port2.gamepad.x",
                "Y": "snes.input.port2.gamepad.y",
                "L": "snes.input.port2.gamepad.l",
                "R": "snes.input.port2.gamepad.r",
                "UP": "snes.input.port2.gamepad.up",
                "DOWN": "snes.input.port2.gamepad.down",
                "LEFT": "snes.input.port2.gamepad.left",
                "RIGHT": "snes.input.port2.gamepad.right",
                "SELECT": "snes.input.port2.gamepad.select",
                "START": "snes.input.port2.gamepad.start",
            }

    def mednafen_system_prefix(self):
        return "snes"

    # def mednafen_video_size(self):
    #     if self.is_pal():
    #         size = (256, 239)
    #     else:
    #         size = (256, 224)
    #     return size



'''OpenGL extension VERSION.GL_4_1

This module customises the behaviour of the 
OpenGL.raw.GL.VERSION.GL_4_1 to provide a more 
Python-friendly API

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/VERSION/GL_4_1.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
from OpenGL.GL import glget
import ctypes
from OpenGL.raw.GL import _types
from OpenGL.raw.GL.VERSION.GL_4_1 import *
from OpenGL.raw.GL.VERSION.GL_4_1 import _EXTENSION_NAME

def glInitGl41VERSION():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

### END AUTOGENERATED SECTION
from OpenGL.GL.ARB.ES2_compatibility import *
from OpenGL.GL.ARB.get_program_binary import *
from OpenGL.GL.ARB.separate_shader_objects import *
from OpenGL.GL.ARB.shader_precision import *
from OpenGL.GL.ARB.vertex_attrib_64bit import *
from OpenGL.GL.ARB.viewport_array import *



from fs_uae_launcher.ui.settings.OptionUI import OptionUI
import fsui
from fsui.extra.iconheader import NewIconHeader
from fs_uae_launcher.res import gettext
from ...Settings import Settings


LANGUAGE_ITEMS = [
    # language name, language code, country flag code
    (gettext("Automatic"), "", "unknown"),
    ("Čeština", "cs", "cz"),
    ("Dansk", "da", "dk"),
    ("Deutsch", "de", "de"),
    ("English", "en", "gb"),
    ("Español", "es", "es"),
    ("Français", "fr", "fr"),
    ("Eλληνικά", "el", "gr"),  # greek
    ("Italiano", "it", "it"),
    ("Norsk (bokmål)", "nb", "no"),
    ("Polski", "pl", "pl"),
    ("Português", "pt", "pt"),
    ("Srpski", "sr", "rs"),  # srpski or српски
    ("Svensk", "sv", "se"),
    ("Suomi", "fi", "fi"),
    ("Türkçe", "tr", "tr"),
]


class LanguageSettingsPage(fsui.Panel):

    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        self.layout = fsui.VerticalLayout()
        # self.layout.set_padding(20, 20, 20, 20)

        self.icon_header = NewIconHeader(
            self, fsui.Icon("language-settings", "pkg:fs_uae_workspace"),
            # gettext("Language Preference"),
            gettext("Appearance"),
            gettext("Set language and look for FS-UAE applications"))
        self.layout.add(self.icon_header, fill=True, margin_bottom=20)

        # label = fsui.Label(self, "Select language:")
        # self.layout.add(label, fill=True)
        # self.layout.add_spacer(6)

        def add_option(name):
            self.layout.add(OptionUI.create_group(self, name), fill=True,
                            margin_top=10, margin_bottom=10)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        # hori_layout.add(fsui.Label(self, gettext("Language:")))
        hori_layout.add(fsui.Label(self, gettext("Language:")))
        hori_layout.add_spacer(0, expand=True)
        self.language_choice = LanguageSettingChoice(self)
        hori_layout.add(self.language_choice)

        self.layout.add_spacer(20)

        information = ""
        # information = gettext(
        #     "A change of language will only affect applications "
        #     "which are restarted after the change.")
        # information += "\n\n"
        information += gettext(
            "When Automatic is specified, your preferred language is set "
            "based on information from the operating system (or English, "
            "if a supported language is not detected).")
        self.layout.add(
            fsui.MultiLineLabel(self, information, 640))

        label = fsui.HeadingLabel(self, "FS-UAE Launcher")
        self.layout.add(label, margin_top=20, margin_bottom=20)

        add_option("launcher_theme")
        add_option("launcher_font_size")

        self.layout.add_spacer(10)

        # self.set_size(self.layout.get_min_size())
        # self.set_size((400, 400))

    def __del__(self):
        print("LanguageWindow.__del__")


class LanguageSettingChoice(fsui.Choice):

    def __init__(self, parent):
        fsui.Choice.__init__(self, parent)
        # FIXME: include in () what language is currently the automatic one

        selected_language = Settings.get("language")
        k = 0
        selected_index = 0
        for label, language_key, icon_name in LANGUAGE_ITEMS:
            self.add_item(label, fsui.Image(
                "fs_uae_workspace:res/16/flag-{0}.png".format(icon_name)))
            if language_key == selected_language:
                selected_index = k
            k += 1
        if selected_index > 0:
            self.set_index(selected_index)

        self.item_selected.connect(self.__item_changed)

    def __item_changed(self, index):
        Settings.set("language", LANGUAGE_ITEMS[index][1])



import fsui
from ...I18N import gettext
from .CustomOptionsPage import CustomOptionsPage


class ConfigDialog(fsui.Dialog):

    def __init__(self, parent):
        super().__init__(parent, gettext("Custom Configuration"))
        buttons, layout = fsui.DialogButtons.create_with_layout(self)
        buttons.create_close_button()
        layout.add(CustomOptionsPage(self), fill=True, expand=True)

    @classmethod
    def run(cls, parent):
        cls(parent).show()



from fsgs.platform import PlatformHandler
from fsgs.mednafen.game_boy import GameBoyRunner
from .loader import SimpleLoader


class GameBoyPlatformHandler(PlatformHandler):

    PLATFORM_NAME = "Game Boy"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return GameBoyLoader(fsgs)

    def get_runner(self, fsgs):
        return GameBoyRunner(fsgs)


class GameBoyLoader(SimpleLoader):
    pass



import os
from fsbc.desktop import set_open_url_in_browser_function
from fsbc.system import macosx
from fsui.qt import QApplication, QDesktopServices, QUrl
from fsbc.Application import Application as BaseApplication


def open_url_in_browser(url):
    print("[QT] open_url_in_browser", url)
    # noinspection PyCallByClass,PyTypeChecker,PyArgumentList
    QDesktopServices.openUrl(QUrl(url))


class QtBaseApplication(QApplication):
    pass


# def fix_qt_for_maverick():
#     """ Fixes Qt 4 for Mac OS X 10.9.
#     http://successfulsoftware.net/2013/10/23/fixing-qt-4-for-mac-os-x-10-9-mavericks/
#     """
#     if not macosx:
#         return
#     #if QSysInfo.MacintoshVersion <= QSysInfo.MV_10_8:
#     #    return
#
#     # fix Mac OS X 10.9 (mavericks) font issue
#     # https://bugreports.qt-project.org/browse/QTBUG-32789
#     QFont.insertSubstitution(".Lucida Grande UI", "Lucida Grande")


class Application(BaseApplication):

    def __init__(self, name):
        BaseApplication.__init__(self, name)

        if macosx:
            # qt_plugins_dir = os.path.join(
            #     BaseApplication.executable_dir(), "..", "Resources",
            #     "qt_plugins")
            # print(qt_plugins_dir)
            # if os.path.exists(qt_plugins_dir):
            #     QApplication.setLibraryPaths([qt_plugins_dir])
            if os.path.exists(os.path.join(BaseApplication.executable_dir(),
                                           "platforms", "libqcocoa.dylib")):
                # noinspection PyCallByClass,PyTypeChecker,PyArgumentList
                QApplication.setLibraryPaths(
                    [BaseApplication.executable_dir()])

        # should not be necessary with Qt 5.2.x:
        # fix_qt_for_maverick()
        set_open_url_in_browser_function(open_url_in_browser)
        self.qapplication = QtBaseApplication([])
        # self.qapplication.setApplicationName("FS-UAE Launcher")

        self.on_create()

    def on_create(self):
        pass

    def run(self):
        self.qapplication.exec_()
        self.stop()

    def set_icon(self, icon):
        self.qapplication.setWindowIcon(icon.qicon())

    def run_in_main(self, function, *args, **kwargs):
        from fsui.qt import call_after
        call_after(function, *args, **kwargs)



try:
    from Xlib.display import Display
except ImportError:
    Display = None


def configure_dock_window(widget, left, right, top, bottom):
    xid = widget.winId()
    # convert to int in case it is a sip.voidptr object (pyqt5)
    xid = int(xid)
    display = Display()
    window = display.create_resource_object("window", xid)
    # _NET_WM_DESKTOP = display.intern_atom("_NET_WM_DESKTOP")
    _NET_WM_STRUT = display.intern_atom("_NET_WM_STRUT")
    _NET_WM_STATE = display.intern_atom("_NET_WM_STATE")
    # noinspection PyPep8Naming
    ATOM = display.intern_atom("ATOM")
    # _NET_WM_STATE_SKIP_TASKBAR = display.intern_atom(
    #     "_NET_WM_STATE_SKIP_TASKBAR")
    # _NET_WM_STATE_SKIP_PAGER = display.intern_atom(
    #     "_NET_WM_STATE_SKIP_PAGER")
    _NET_WM_STATE_BELOW = display.intern_atom(
        "_NET_WM_STATE_BELOW")
    # _NET_WM_USER_TIME = display.intern_atom(
    #     "_NET_WM_USER_TIME")
    _NET_WM_WINDOW_TYPE = display.intern_atom(
        "_NET_WM_WINDOW_TYPE")
    _NET_WM_WINDOW_TYPE_DOCK = display.intern_atom(
        "_NET_WM_WINDOW_TYPE_DOCK")
    # noinspection PyPep8Naming
    CARDINAL = display.intern_atom("CARDINAL")

    # Set window type to dock (skip task bar, no focus, on all desktops)
    window.change_property(
        _NET_WM_WINDOW_TYPE, ATOM, 32, [_NET_WM_WINDOW_TYPE_DOCK])
    # Reserve space for title bar and task bar
    window.change_property(
        _NET_WM_STRUT, CARDINAL, 32, [left, right, top, bottom])

    # Show on all workspaces
    # window.change_property(
    #     _NET_WM_DESKTOP, CARDINAL, 32, [0xffffffff, 0, 0, 0])
    # Don't focus the window on window creation
    # window.change_property(
    #     _NET_WM_USER_TIME, CARDINAL, 32, [0, 0, 0, 0])
    # Don't show in task bar or window switcher
    # window.change_property(
    #     _NET_WM_STATE, ATOM, 32,
    #     [_NET_WM_STATE_SKIP_TASKBAR, _NET_WM_STATE_SKIP_PAGER])

    #
    window.change_property(_NET_WM_STATE, ATOM, 32, [_NET_WM_STATE_BELOW])
    display.sync()



import fsui as fsui
from ..I18N import gettext


class LaunchDialog(fsui.LegacyDialog):

    def __init__(self, parent, title, task):
        fsui.LegacyDialog.__init__(self, parent, title)
        self.layout = fsui.VerticalLayout()

        self.layout.add_spacer(400, 20)

        hor_layout = fsui.HorizontalLayout()
        self.layout.add(hor_layout, fill=True)

        hor_layout.padding_right = 20
        hor_layout.add_spacer(20)

        image = fsui.Image("fs_uae_launcher:res/fs_uae_group.png")
        self.image_view = fsui.ImageView(self, image)
        hor_layout.add(self.image_view, valign=0.0)
        hor_layout.add_spacer(20)

        ver_layout = fsui.VerticalLayout()
        hor_layout.add(ver_layout, fill=True, expand=True)
        self.title_label = fsui.HeadingLabel(self, title)
        ver_layout.add(self.title_label, fill=True)

        ver_layout.add_spacer(6)
        self.sub_title_label = fsui.Label(self, gettext("Preparing..."))
        ver_layout.add(self.sub_title_label, fill=True)

        self.layout.add_spacer(20)

        hor_layout = fsui.HorizontalLayout()
        self.layout.add(hor_layout, fill=True)

        hor_layout.add_spacer(20, expand=True)
        self.cancel_button = fsui.Button(self, gettext("Cancel"))
        self.cancel_button.activated.connect(self.on_cancel_button)
        hor_layout.add(self.cancel_button)
        hor_layout.add_spacer(20)

        self.layout.add_spacer(20)
        self.set_size(self.layout.get_min_size())
        self.center_on_parent()

        self.was_closed = False
        self.task = task
        self.task.progressed.connect(self.on_progress)
        self.task.finished.connect(self.on_complete)
        self.task.failed.connect(self.on_error)

    def complete(self):
        self.was_closed = True
        self.end_modal(0)

    def on_close(self):
        self.cancel()
        return False

    def on_progress(self, progress):
        def function():
            self.sub_title_label.set_text(progress)
        fsui.call_after(function)

    def on_complete(self):
        def function():
            self.complete()
        fsui.call_after(function)

    def on_error(self, message):
        self.end_modal(1)
        fsui.show_error(message)

    def on_cancel_button(self):
        self.cancel()
        # self.handler.on_progress = None
        # self.handler.on_complete = None
        # self.complete()

    def cancel(self):
        print("FIXME: LaunchDialog.cancel")
        self.task.stop()
        self.cancel_button.disable()

    # def handler_thread(self):
    #     try:
    #         self._handler_thread()
    #     except Exception:
    #         traceback.print_exc()
    #         message = traceback.format_exc()
    #
    #         def function():
    #             self.on_error(message)
    #
    #         fsui.call_after(function)
    #
    # def _handler_thread(self):
    #     self.handler.run_sequence()



import os
import time
from fsbc.Application import app
from game_center.gamecentersettings import GameCenterSettings
from game_center.glui.constants import TOP_ITEM_LEFT
from game_center.glui.constants import TOP_ITEM_NOBORDER
from .items import MenuItem
from game_center.glui.navigatable import Navigatable
from .opengl import fs_emu_blending
from game_center.glui.state import State
from game_center.glui.sdl import *
from game_center.glui.texture import Texture


def post_quit_event():
    # pygame.event.post(pygame.event.Event(pygame.QUIT))
    # app.quit()
    State.quit = True


class TopMenuItem(MenuItem):

    def update_size_right(self):
        self.w = 80

    def render_top_right(self, selected=False):
        self.render_top_background(selected, style=TOP_ITEM_LEFT)
        fs_emu_blending(True)
        if selected:
            texture = self.selected_texture
        else:
            texture = self.normal_texture
        texture.render(self.x, self.y, self.w, self.h)


class CloseItem(TopMenuItem):
    def __init__(self):
        TopMenuItem.__init__(self)
        self.normal_texture = Texture.close
        self.selected_texture = Texture.close_selected

    def activate(self, menu):
        post_quit_event()


class ShutdownItem(TopMenuItem):
    def __init__(self):
        TopMenuItem.__init__(self)
        self.normal_texture = Texture.shutdown
        self.selected_texture = Texture.shutdown_selected

    def activate(self, menu):
        command = GameCenterSettings.get_shutdown_command()
        if command:
            os.system(command)
        post_quit_event()


class MaximizeItem(TopMenuItem):
    def __init__(self):
        MenuItem.__init__(self)
        self.title = "Ma"


class MinimizeItem(TopMenuItem):
    def __init__(self):
        MenuItem.__init__(self)
        self.normal_texture = Texture.minimize
        self.selected_texture = Texture.minimize_selected

    def activate(self, menu):
        SDL_Minimize()
        # want item menu focused when window is restored
        # FIXME: temporarily disabled
        # set_current_menu(State.current_menu)


class ClockItem(MenuItem):

    def __init__(self):
        MenuItem.__init__(self)
        self.title = "00:00"
        self.enabled = False

    def update_size(self, text):
        # w, h = BitmappedFont.title_font.measure(text)
        # self.w = w + 40 # + 8
        # self.w = 134
        self.w = 148
        # print(self.w)

    def render_top_right(self, selected=False):
        self.render_top_background(selected, style=TOP_ITEM_NOBORDER)
        self.render_top(self.get_top_right_text(), selected, right_align=True)

    def get_top_right_text(self):
        return time.strftime("%H:%M")

    def activate(self, menu):
        pass


class TopMenu(Navigatable):

    def __init__(self):
        self.left = []
        self.right = []
        if self.use_clock_item():
            self.right.append(ClockItem())
        # if Render.allow_minimize:
        #     if Config.get_bool("menu/minimize", True):
        #         self.right.append(MinimizeItem())
        command = GameCenterSettings.get_shutdown_command()
        if command:
            self.right.append(ShutdownItem())
        else:
            self.right.append(CloseItem())
        self._selected_index = 0

    def use_clock_item(self):
        return app.settings["game-center:top-clock"] != "0"

    def append_left(self, item):
        self.left.append(item)

    def append_right(self, item):
        self.right.insert(0, item)

    @property
    def selected_item(self):
        index = self._selected_index
        return self[index]

    def get_selected_index(self):
        return self._selected_index

    def set_selected_index(self, index, immediate=False):
        assert immediate is not None
        self._selected_index = index

    def go_left(self, count=1):
        index = self._selected_index - 1
        while index >= 0:
            if self[index].enabled:
                self._selected_index = index
                break
            index -= 1

    def go_right(self, count=1):
        index = min(len(self) - 1, self._selected_index + count)
        self._selected_index = index
        if not self[index].enabled:
            self.go_right()

    def go_up(self):
        # do nothing
        pass

    def activate(self):
        result = self.selected_item.activate(State.current_menu)
        # FIXME:
        from game_center.glui.menu import Menu
        if isinstance(result, Menu):
            # FIXME:
            from game_center.glui.window import enter_menu
            enter_menu(result)

    def __getitem__(self, index):
        if index < len(self.left):
            return self.left[index]
        return self.right[index - len(self.left)]

    def __len__(self):
        return len(self.left) + len(self.right)



import fsui as fsui
from .imports.ImportGroup import ImportGroup
from .ScanKickstartGroup import ScanKickstartGroup


class SetupPanel(fsui.Panel):

    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        self.layout = fsui.VerticalLayout()

        self.scan_kickstart_group = ScanKickstartGroup(self)
        self.import_rom_group = ImportGroup(self)
        self.import_af_group = ImportGroup(self, ImportGroup.AMIGA_FOREVER)

        self.layout.add(self.import_rom_group, fill=True)
        self.layout.add_spacer(20)
        self.layout.add(self.import_af_group, fill=True)
        self.layout.add_spacer(20)
        self.layout.add(self.scan_kickstart_group, fill=True)
        # self.layout.add_spacer(Skin.EXTRA_GROUP_MARGIN)



import fsui as fsui
from ...I18N import gettext
from .ConfigCheckBox import ConfigCheckBox
from ..HelpButton import HelpButton


class ExpansionsGroup(fsui.Group):

    def __init__(self, parent):
        fsui.Group.__init__(self, parent)
        self.layout = fsui.VerticalLayout()

        heading_label = fsui.HeadingLabel(self, gettext("Expansions"))
        self.layout.add(heading_label, margin=10)
        self.layout.add_spacer(0)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)

        vert_layout = fsui.VerticalLayout()
        hori_layout.add(vert_layout, fill=True, expand=True)

        hor2_layout = fsui.HorizontalLayout()
        widget = ConfigCheckBox(
            self, gettext("UAE Graphics Card (Picasso96 Support)"),
            "uaegfx_card")
        # widget.set_tooltip(gettext("Picasso96 Support (uaegfx.card)"))
        hor2_layout.add(widget, expand=True)
        widget = HelpButton(self, "http://fs-uae.net/options#uaegfx-card")
        hor2_layout.add(widget, margin_left=10)
        vert_layout.add(hor2_layout, fill=True, margin=10)

        # vert_layout = fsui.VerticalLayout()
        # hori_layout.add(vert_layout, fill=True, expand=True)

        hor2_layout = fsui.HorizontalLayout()
        # gettext("Built-in TCP/IP Stack")
        widget = ConfigCheckBox(self, "UAE bsdsocket.library",
                                "bsdsocket_library")
        # widget.set_tooltip(
        #     gettext("Built-in TCP/IP Stack (bsdsocket.library)"))
        hor2_layout.add(widget, expand=True)
        widget = HelpButton(
            self, "http://fs-uae.net/options#bsdsocket-library")
        hor2_layout.add(widget, margin_left=10)
        vert_layout.add(hor2_layout, fill=True, margin=10)



import weakref
from fsui.qt import QSignal, Qt, QRect, QEvent
from fsui.qt import QWidget, QVBoxLayout, QColor, QPainter, QPen, QBrush
from fs_uae_workspace.desktop import get_root_window


class WindowContent(QWidget):

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

        self.setAutoFillBackground(True)
        p = self.palette()
        p.setColor(self.backgroundRole(), QColor(0xed, 0xed, 0xed))
        # p.setColor(self.backgroundRole(), QColor(0xff, 0xaa, 0xaa))
        self.setPalette(p)


# class FSUAEDesktopWindow(QWidget):
#
#     def __init__(self, parent=None, title=""):
#         if parent is None:
#             parent = get_root_window()
#         QWidget.__init__(self, parent)
#         self.left_padding = 6
#         self.right_padding = 6
#         self.top_padding = 30
#         self.bottom_padding = 6
#
#         self.setAutoFillBackground(True)
#         p = self.palette()
#         p.setColor(self.backgroundRole(), QColor(0xff, 0xaa, 0xaa))
#         self.setPalette(p)
#
#         # self.__content = WindowContent(self)
#         # layout = QVBoxLayout()
#         # layout.addWidget(self.__content, 1)
#         # self.setLayout(layout)
#         # return self.__content
#
#     def width(self):
#         return QWidget.width(self) - self.left_padding - self.right_padding
#
#     def height(self):
#         return QWidget.height(self) - self.top_padding - self.bottom_padding
#
#     def size(self):
#         return self.width(), self.height()
#
#     def resize(self, w, h):
#         #self.SetClientSize(size)
#         #print("FIXME:\n\nDialog.set_size")
#         QWidget.resize(self, w + self.left_padding + self.right_padding,
#                        h + self.top_padding + self.bottom_padding)
#
#     # def resizeEvent(self, event):
#     #     print("resized..")
#     #     self.on_resize()

class QContent(QWidget):

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

    def mousePressEvent(self, event):
        pass

    def mouseReleaseEvent(self, event):
        pass

    def mouseMoveEvent(self, event):
        pass


class QWindow(QWidget):

    closed = QSignal()

    def __init__(self, parent=None):
        if parent is None:
            parent = get_root_window()

        self._window = weakref.ref(self)

        self.padding = [6, 26, 6, 6]

        self._position_specified = False
        self._size_specified = False

        QWidget.__init__(self, parent)
        self.setAttribute(Qt.WA_DeleteOnClose, True)
        # self.setWindowTitle(title)

        self.setAutoFillBackground(True)
        p = self.palette()
        p.setColor(self.backgroundRole(), QColor(0x739cda))
        # p.setColor(self.backgroundRole(), QColor(0xff, 0xaa, 0xaa))
        self.setPalette(p)

        self.mouse_press_event = None

        self.w = QContent()

        self.w.setAutoFillBackground(True)
        p = self.w.palette()
        p.setColor(self.w.backgroundRole(), QColor(0xededed))
        self.w.setPalette(p)

        layout = QVBoxLayout()
        layout.setContentsMargins(*self.padding)
        layout.addWidget(self.w)
        QWidget.setLayout(self, layout)

        # self.closed = Signal()
        # self.destroyed = Signal()

        self.w.installEventFilter(self)
        # self.installEventFilter(self)
        self.mouse_press_window_pos = None

    def eventFilter(self, obj, event):
        assert isinstance(event, QEvent)
        if event.type() == QEvent.MouseButtonPress:
            self.raise_()
        return False

    def set_position(self, position):
        self.move(position[0], position[1])

    def move(self, x, y):
        QWidget.move(self, x, y)
        self._position_specified = True

    def resize(self, w, h):
        w += self.padding[0] + self.padding[2]
        h += self.padding[1] + self.padding[3]
        QWidget.resize(self, w, h)

    def show(self):
        if not self._position_specified:
            from fs_uae_workspace.shell import set_initial_window_position
            set_initial_window_position(self)
        QWidget.show(self)

    def is_shown(self):
        return self.isVisible()

    def setLayout(self, layout):
        self.w.setLayout(layout)

    # noinspection PyPep8Naming
    def setCentralWidget(self, widget):

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(widget)
        self.w.setLayout(layout)

    def mousePressEvent(self, event):
        if True:
            self.raise_()
        # mouse is automatically grabbed by QT, as long as the mouse button
        # is pressed, so no need to grab it here.
        self.mouse_press_event = (event.x(), event.y(), event.globalX(),
                                  event.globalY())
        self.mouse_press_window_pos = self.pos()

    def mouseReleaseEvent(self, event):
        if self.mouse_press_event[0] < 26 and self.mouse_press_event[1] < \
                26 and event.x() < 26 and event.y() < 26:
            print("close")
            self.close()
        # print("mouseReleaseEvent")
        self.mouse_press_event = None

    def mouseMoveEvent(self, event):
        if self.mouse_press_event is not None:
            dx = event.globalX() - self.mouse_press_event[2]
            dy = event.globalY() - self.mouse_press_event[3]
            x = self.mouse_press_window_pos.x() + dx
            y = self.mouse_press_window_pos.y() + dy
            # self.move(QPoint(x, y))
            self.move(x, y)

    def paintEvent(self, event):
        painter = QPainter(self)
        # painter.setRenderHints(QPainter.Antialiasing)

        pen = QPen(QColor(0x0, 0x0, 0x0))
        pen.setWidth(1)
        painter.setPen(pen)
        painter.setBrush(QBrush(Qt.white))
        rect = QRect(9, 9, 8, 8)
        painter.drawRect(rect)

        painter.drawText(
            QRect(24, 0, 1000, 26), Qt.AlignVCenter, self.windowTitle())

        painter.setPen(QPen(QColor(0xacccfb)))
        painter.drawLine(0, 0, self.width(), 0)
        painter.drawLine(0, 1, 0, self.height())

        painter.setPen(QPen(QColor(0x5a7cae)))
        painter.drawLine(self.width() - 1, 0, self.width() - 1, self.height())
        painter.drawLine(0, self.height() - 1, self.width(), self.height() - 1)

    def closeEvent(self, event):
        print("Window.closeEvent")
        # for function in self.close_listeners:
        #     function()
        self.closed.emit()
        self.on_close()
        event.accept()
        # self.destroy()
        # event.ignore()

    def on_close(self):
        pass


class Window(QWindow):

    def __init__(self, parent=None, title=""):
        QWindow.__init__(self, parent)
        self.setWindowTitle(title)
        self.layout = None

        self._container = WindowContent(self)
        self.setCentralWidget(self._container)

        self.destroy_listeners = []
        self.close_listeners = []

    def get_container(self):
        return self._container

    def get_parent(self):
        return None

    def add_destroy_listener(self, function):
        self.destroy_listeners.append(function)

    def add_close_listener(self, function):
        # self.close_listeners.append(function)
        self.closed.connect(function)

    def get_window(self):
        return self

    def on_destroy(self):
        pass

    def __destroyed(self):
        print("__destroyed")
        for function in self.destroy_listeners:
            function()
        self.on_destroy()

    def showEvent(self, event):
        print("showed..")
        self.on_resize()

    def get_position(self):
        return self.pos().x(), self.pos().y()

    def get_size(self):
        w = self.width()
        h = self.height()
        w -= self.padding[0] + self.padding[2]
        h -= self.padding[1] + self.padding[3]
        return w, h

    def set_size(self, size):
        w, h = size
        self._size_specified = True
        self.resize(w, h)

    def is_maximized(self):
        # return self.isMaximized()
        print("FIXME: always returning False")
        return False

    def maximize(self):
        self.showMaximized()

    def center_on_parent(self):
        # self.CenterOnParent()
        print("FIXME:\n\nWindow.center_on_parent")

    def center_on_screen(self):
        # self.CenterOnParent()
        print("FIXME:\n\nWindow.center_on_screen")

    def resizeEvent(self, event):
        print("resized..")
        self.on_resize()

    def on_resize(self):
        if self.layout:
            self.layout.set_size(self.get_size())
            self.layout.update()

    def set_background_color(self, color):
        print("FIXME: Window.set_background_color")

    def set_icon_from_path(self, path):
        print("FIXME: Window.set_icon_from_path")

    def show(self):
        if hasattr(self, "layout") and not self._size_specified:
            self.set_size(self.layout.get_min_size())
        QWindow.show(self)



from fsbc.Application import app
import fsui as fsui
from ...Config import Config
from ...Settings import Settings


class CustomSettingsPage(fsui.Panel):

    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        self.layout = fsui.VerticalLayout()

        # self.layout.add_spacer(580, 20)

        hor_layout = fsui.HorizontalLayout()
        self.layout.add(hor_layout, fill=True, expand=True)

        # hor_layout.add_spacer(20)
        self.text_area = fsui.TextArea(self, font_family="monospace")
        # self.text_area.set_min_height(400)
        self.text_area.set_text(self.get_initial_text())
        hor_layout.add(self.text_area, fill=True, expand=True)

        # hor_layout.add_spacer(20)

        # self.layout.add_spacer(20)

        # hor_layout = fsui.HorizontalLayout()
        # self.layout.add(hor_layout, fill=True)

        # hor_layout.add_spacer(20, expand=True)
        # self.close_button = fsui.Button(self, _("Close"))
        # self.close_button.activated.connect(self.on_close_button)
        # hor_layout.add(self.close_button)
        # hor_layout.add_spacer(20)

        # self.layout.add_spacer(20)
        # self.set_size(self.layout.get_min_size())
        # self.center_on_parent()

        # self.get_window().add_close_listener(self.on_close_window)

        self.text_area.changed.connect(self.update_settings)

    # def on_close_window(self):
    #     self.update_settings()

    # def on_close_button(self):
    #     self.end_modal(0)

    def update_settings(self):
        text = self.text_area.get_text()
        # FIXME: accessing values directly here, not very nice
        keys = list(app.settings.values.keys())
        for key in keys:
            if key not in Settings.default_settings:
                Settings.set(key, "")
                del app.settings.values[key]

        for line in text.split("\n"):
            line = line.strip()
            if line.startswith("# You can write key = value pairs here"):
                continue
            parts = line.split("=", 1)
            if len(parts) == 2:
                key = parts[0].strip()
                # if key in Settings.default_settings:
                #     continue
                value = parts[1].strip()
                app.settings[key] = value

    def get_initial_text(self):
        text = DEFAULT_TEXT
        # FIXME: accessing values directly here, not very nice
        keys = app.settings.values.keys()
        for key in sorted(keys):
            if key in Settings.default_settings:
                continue
            #    #print("(settings) ignoring key", key)
            #    text += "# key {0} will be ignored\n".format(key)
            # if key in Config.config_keys:
            #     print("(settings) ignoring key", key)
            #     continue
            if key in Config.config_keys:
                # print("(settings) ignoring key", key)
                text += "\n# {0} is ignored here " \
                        "(use config dialog instead)\n".format(key)
            value = app.settings[key]
            if Config.get(key):
                text += "\n# {0} is overridden by current " \
                        "configuration\n".format(key)
            text += "{0} = {1}\n".format(key, value)
            if Config.get(key):
                text += "\n"
            if key in Config.config_keys:
                text += "\n"
        return text

DEFAULT_TEXT = """\
# You can write key = value pairs here to set FS-UAE options
# not currently supported by the user interface. This is only a
# temporary feature until the GUI supports all options directly.
#
# The options specified here are global and will apply to all
# configurations. Config options such as hardware and memory
# options will be ignored. Options suitable here are options
# like theme options.

"""



from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os
import io
import xml.etree.cElementTree as ElementTree


class DatFile(object):
    
    def __init__(self, file=None):
        self.games = []
        self.description = ""

        self.extensions = []
        if file is not None:
            self.load(file)

    def reset(self):
        self.games = []
        self.description = ""

    def load(self, file):
        if hasattr(file, "read"):
            self._load_file(file)
        else:
            self._load_path(file)

    def _load_path(self, path):
        with io.open(path, "r", encoding="UTF-8") as f:
            self._load_file(f)

    def _load_file(self, f):
        self.load_data(f.read())

    def load_data(self, data):
        if data.startswith("<") or data.startswith("\uFEFF"):
            # assume XML
            self._load_xml(data)
        else:
            self._load_dat(data)

    def _load_xml(self, data):
        self.reset()

        root = ElementTree.fromstring(data)
        header_node = root.find("header")
        self.description = header_node.find("description").text.strip()
        for game_node in root.findall("game"):
            game = {
                "name": game_node.attrib["name"],
                "files": [],
            }
            for rom_node in game_node.findall("rom"):
                rom = {"name": rom_node.attrib["name"],
                       "size": int(rom_node.attrib["size"]),
                       "crc32": rom_node.attrib["crc"],
                       "md5": rom_node.attrib["md5"],
                       "sha1": rom_node.attrib["sha1"]}
                game["files"].append(rom)
            self.games.append(game)

    def _load_dat(self, data):
        self.reset()

        # this is a very quick-and-dirty/hackish parser...
        game = None
        for line in data.split("\n"):
            line = line.strip()
            if line.startswith("game ("):
                game = {
                    "name": "",
                    "files": [],
                }
                self.games.append(game)
            if line.startswith("name "):
                if game is not None:
                    if '"' in line:
                        parts = line.split('"')
                        assert len(parts) == 3
                        game["name"] = parts[1]
                    else:
                        game["name"] = line[5:]
            elif line.startswith("description "):
                if game is None:
                    parts = line.split('"')
                    assert len(parts) == 3
                    self.description = parts[1]
            elif line.startswith("rom ("):
                rom = {}
                parts = line.split('"')
                if len(parts) == 3:
                    rom["name"] = parts[1]
                else:
                    parts = line.split(' ')
                    rom["name"] = parts[3]
                n, ext = os.path.splitext(rom["name"])
                if self.extensions and ext.lower() not in self.extensions:
                    continue
                parts = line.split(" size ")
                rom["size"] = int(parts[1].strip().split(" ")[0].strip())
                parts = line.split(" crc ")
                rom["crc32"] = parts[1].strip().split(" ")[0].strip()
                parts = line.split(" md5 ")
                rom["md5"] = parts[1].strip().split(" ")[0].strip()
                parts = line.split(" sha1 ")
                rom["sha1"] = parts[1].strip().split(" ")[0].strip()
                game["files"].append(rom)



import os
import shutil
from fsbc.fs import get_data_dir
from fsgs.FSGSDirectories import FSGSDirectories


class ApplicationMixin(object):

    def get_settings_path(self):
        settings_ini = os.path.join(FSGSDirectories.get_data_dir(),
                                    "Settings.ini")
        if not os.path.exists(settings_ini):
            migrate_list = [os.path.join(FSGSDirectories.get_data_dir(),
                                         "FS-UAE Launcher.ini")]
            portable_ini = os.path.join(FSGSDirectories.get_base_dir(),
                                        "Portable.ini")
            if not os.path.exists(portable_ini):
                # not portable, can migrate settings from old
                # launcher.settings file
                migrate_list.append(os.path.join(
                    get_data_dir(), "fs-uae", "launcher.settings"))
            # move the highest-priority settings file if present
            for migrate in migrate_list:
                if os.path.exists(migrate):
                    shutil.move(migrate, settings_ini)
                    break
            # remove all old settings files
            for migrate in migrate_list:
                if os.path.exists(migrate):
                    os.remove(migrate)

        return settings_ini



from fs_uae_launcher.ui.ConfigGroup import ConfigGroup
from fs_uae_launcher.ui.bottombar.RatingButton import RatingButton
from fs_uae_launcher.ui.newbutton import NewButton
from fs_uae_workspace.shell import shell_open
import fsui as fsui
from ..Settings import Settings
from .ConfigurationsBrowser import ConfigurationsBrowser
from .GameListSelector import GameListSelector
from .Skin import Skin
from .VariantsBrowser import VariantsBrowser


class ConfigurationsPanel(fsui.Panel):

    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        Skin.set_background_color(self)
        self.layout = fsui.HorizontalLayout()

        vert_layout = fsui.VerticalLayout()
        self.layout.add(vert_layout, fill=True, expand=True)

        hor_layout = fsui.HorizontalLayout()
        vert_layout.add(hor_layout, fill=True)

        label_stand_in = fsui.Panel(self)
        tw, th = label_stand_in.measure_text("Games")
        label_stand_in.set_min_height(th)
        hor_layout.add(label_stand_in, margin_top=10, margin_bottom=10)

        hor_layout.add(NewButton(self), margin_left=10, margin_right=10)

        game_list_selector = GameListSelector(self)
        game_list_selector.set_min_width(250)
        game_list_selector.setMaximumWidth(250)
        hor_layout.add(game_list_selector, expand=False, margin_left=10)

        self.text_field = fsui.TextField(self, Settings.get("config_search"))
        self.text_field.on_change = self.on_search_change
        if VariantsBrowser.use_horizontal_layout():
            # window is big enough to use fixed size
            # self.text_field.set_min_width(210)
            self.text_field.set_min_width(229)
            hor_layout.add(
                self.text_field, expand=False, margin=10, margin_top=0,
                margin_bottom=0)
        else:
            hor_layout.add(
                self.text_field, expand=True, margin=10, margin_top=0,
                margin_bottom=0)

        # self.refresh_button = IconButton(self, "refresh_button.png")
        # self.refresh_button.set_tooltip(
        #     gettext("Refresh Game Configurations from Online Database"))
        # self.refresh_button.activated.connect(self.on_refresh_button)
        # hor_layout.add(
        #     self.refresh_button, margin=10, margin_top=0, margin_bottom=0)

        self.configurations_browser = ConfigurationsBrowser(self)

        vert_layout.add(
            self.configurations_browser, fill=True, expand=3, margin=10)

        self.variants_panel = fsui.Panel(self)
        vert_layout.add(self.variants_panel, fill=True, expand=False,
                        margin=10, margin_top=20)

        self.variants_panel.layout = fsui.HorizontalLayout()
        self.variants_browser = VariantsBrowser(self.variants_panel)
        # Do not use fill=True with the default OS X theme at least,
        # if you do the item will be rendered with the old Aqua look
        self.variants_panel.layout.add(
            self.variants_browser, fill=False, expand=True)

        for rating in [1, 4, 5]:
            button = RatingButton(self.variants_panel, rating)
            self.variants_panel.layout.add(button, margin_left=5, fill=True)

        self.config_panel = fsui.Panel(self)
        vert_layout.add(self.config_panel, fill=True, expand=False,
                        margin_bottom=10, margin_top=20)

        self.config_panel.layout = fsui.VerticalLayout()
        self.config_group = ConfigGroup(self.config_panel, new_button=False)
        self.config_panel.layout.add(self.config_group, fill=True, expand=True)

        Settings.add_listener(self)
        self.on_setting("parent_uuid", Settings.get("parent_uuid"))

    def on_destroy(self):
        Settings.remove_listener(self)

    def on_setting(self, key, value):
        if key == "parent_uuid":
            self.variants_panel.show_or_hide(bool(value))
            self.config_panel.show_or_hide(not bool(value))
            self.layout.update()

    def on_verified_button(self):
        pass

    def on_favorite_button(self):
        pass

    def on_refresh_button(self):
        print("on_refresh_button")
        shell_open("Workspace:Tools/Refresh", parent=self.get_window())

    def on_search_change(self):
        text = self.text_field.get_text()
        Settings.set("config_search", text)



from fsgs.mednafen.mednafen import MednafenRunner


class MasterSystemRunner(MednafenRunner):

    CONTROLLER = {
        "type": "gamepad",
        "description": "Gamepad",
        "mapping_name": "mastersystem",
    }

    PORTS = [
        {
            "description": "1st Controller",
            "types": [CONTROLLER]
        }, {
            "description": "2nd Controller",
            "types": [CONTROLLER]
        },
    ]

    def __init__(self, fsgs):
        super().__init__(fsgs)

    def mednafen_system_prefix(self):
        return "sms"

    def mednafen_input_mapping(self, port):
        if port == 0:
            return {
                "1": "sms.input.port1.gamepad.fire1",
                "2": "sms.input.port1.gamepad.fire2",
                "UP": "sms.input.port1.gamepad.up",
                "DOWN": "sms.input.port1.gamepad.down",
                "LEFT": "sms.input.port1.gamepad.left",
                "RIGHT": "sms.input.port1.gamepad.right",
                "PAUSE": "sms.input.port1.gamepad.pause",
            }
        elif port == 1:
            return {
                "1": "sms.input.port2.gamepad.fire1",
                "2": "sms.input.port2.gamepad.fire2",
                "UP": "sms.input.port2.gamepad.up",
                "DOWN": "sms.input.port2.gamepad.down",
                "LEFT": "sms.input.port2.gamepad.left",
                "RIGHT": "sms.input.port2.gamepad.right",
                "PAUSE": "sms.input.port2.gamepad.pause",
            }

        # FIXME: add PAUSE button to universal gamepad config

    # def mednafen_video_size(self):
    #     return 256, 192



import fsui as fsui
from .TabPanel import TabPanel
from .TabButton import TabButton
from .Constants import Constants
from .Skin import Skin


class WindowWithTabs(fsui.Window):

    def __init__(self, parent, title, border=True):
        fsui.Window.__init__(self, parent, title, border=border)
        Skin.set_background_color(self)
        self.toolbar = None
        self.tab_panel = TabPanel(self)
        self.layout = fsui.VerticalLayout()
        if self.tab_panel:
            self.layout.add(self.tab_panel, fill=True)
        self.current_tab_group_id = 0
        self.tab_groups = [[]]

    # def __size_event(self, event):
    #     size = event.GetSize()
    #     print("WindowWithTabs size event, size =", size)
    #     event.Skip()

    def realize_tabs(self):
        if self.toolbar:
            self.toolbar.Realize()

    def new_tab_group(self):
        self.current_tab_group_id += 1
        self.tab_groups.append([])

    def set_content(self, content):
        self.layout.add(content, expand=True, fill=True)

    def select_tab(self, index, group):
        if self.toolbar:
            pass
        else:
            print("\n\n\nselect tab", index, group)
            # noinspection PyUnresolvedReferences
            self.tab_groups[group][index].select()
            # self.tab_groups[group].select_tab(index)

    def add_tab(self, function, icon, title="", tooltip=""):
        if not tooltip:
            tooltip = title
        button = TabButton(self.tab_panel, icon)
        button.set_tooltip(tooltip)
        button.group_id = self.current_tab_group_id
        button.on_select = function
        self.tab_panel.add(button)
        # noinspection PyTypeChecker
        self.tab_groups[self.current_tab_group_id].append(button)

    def add_tab_button(self, function, icon, title="", tooltip="",
                       menu_function=None, left_padding=0, right_padding=0):
        if not tooltip:
            tooltip = title
        button = TabButton(
            self.tab_panel, icon, type=TabButton.TYPE_BUTTON,
            left_padding=left_padding, right_padding=right_padding)
        button.set_tooltip(tooltip)
        button.group_id = self.current_tab_group_id
        if function:
            button.activated.connect(function)
        elif menu_function:
            def menu_wrapper():
                menu_function()
                button.check_hover()
            button.on_left_down = menu_wrapper
        self.tab_panel.add(button)
        # noinspection PyTypeChecker
        self.tab_groups[self.current_tab_group_id].append(button)
        return button

    def add_tab_panel(self, class_, min_width=0, expand=1000000):
        # panel = class_(self.tab_panel, padding_bottom=2)
        panel = class_(self.tab_panel)
        panel.expandable = True
        panel.set_min_height(Constants.TAB_HEIGHT)
        if self.toolbar:
            panel.SetSize((min_width, 46))
            self.toolbar.AddControl(panel)
        else:
            self.tab_panel.add(panel, expand=expand)
        return panel

    def add_tab_separator(self):
        if self.toolbar:
            self.toolbar.AddSeparator()

    def add_tab_spacer(self, spacer=0, expand=False):
        self.tab_panel.add_spacer(spacer, expand=expand)



from fsgs.platform import PlatformHandler
from fsgs.mednafen.mega_drive import MegaDriveRunner
from .loader import SimpleLoader


class MegaDrivePlatformHandler(PlatformHandler):

    PLATFORM_NAME = "Mega Drive"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return MegaDriveLoader(fsgs)

    def get_runner(self, fsgs):
        return MegaDriveRunner(fsgs)


class MegaDriveLoader(SimpleLoader):
    pass



from fsgs.mame.mame import MAMERunner


class MESSRunner(MAMERunner):

    def is_pal(self):
        # return self.config.get("ntsc_mode") != "1"
        # return False
        refresh_rate = self.get_game_refresh_rate()
        if refresh_rate:
            return int(round(refresh_rate)) == 50

    def mame_emulator_name(self):
        return "mess"

    def mame_init_input(self):
        self.mess_init_input()

    def mess_init_input(self):
        pass

    def mame_romset(self):
        return self.mess_romset()

    def mess_romset(self):
        pass

    # def mess_get_firmware_name(self):
    #     return self.context.game.platform + ' Firmware'

    def mame_prepare(self):
        pass
        # firmware_name = self.mess_get_firmware_name()
        # if firmware_name:
        #     self.mess_firmware_dir = self.prepare_firmware(firmware_name)
        # else:
        #     self.mess_firmware_dir = None

    def mame_configure(self):
        self.add_arg(self.mess_romset()[0])
        # self.configure_mame()
        # with open(self.mednafen_cfg_path(), 'wb') as f:
        #     self._configure_emulator(f)
        # bios_dir = self.get_bios_dir()
        # if bios_dir:
        #     self.args.extend(['-rompath', bios_dir])
        # self.args.extend(['-joystick_deadzone', '0.15'])

        self.mess_configure()
        # self.mess_configure_media()

        if not self.mess_full_keyboard():
            # start mess with UI keyboard keys enabled by default,
            # full keyboard can be activated with INS / Scroll-Lock
            self.add_arg("-ui_active")

    def mess_full_keyboard(self):
        return False

    def mess_configure(self):
        """override in subclasses to provide custom configuration"""
        pass

    def run(self):
        # return self.start_emulator("fs-mess/mess")
        return self.start_emulator_from_plugin_resource("mess")

    def mame_input_mapping(self, port):
        return self.mess_input_mapping(port)

    def mess_input_mapping(self, port):
        pass

    def mame_get_bios_dir(self):
        print("xxxxxxxxx", self.mess_firmware_dir)
        return self.mess_firmware_dir

    def mess_offset_and_scale(self):
        raise NotImplementedError()

    def mame_offset_and_scale(self):
        try:
            return self.mess_offset_and_scale()
        except NotImplementedError:
            return super().mame_offset_and_scale()

#    def mess_configure_media(self):
#        pass

    def mess_configure_cartridge(self, slot="cart"):
        self.add_arg("-{0}".format(slot), self.get_game_file())

    def mess_configure_floppies(self, slots):
        # FIXME: this is a quick hack only..
        self.add_arg("-{0}".format(slots[0]), self.get_game_file())

    def inject_fake_input_string_list(self, delay, inject_string):
        self.set_env("FSGS_FAKE_INPUT_DELAY", str(delay))
        self.set_env("FSGS_FAKE_INPUT", "".join(inject_string))

    def inject_fake_input_string(self, delay, inject_string):

        def inject(inject_key_code):
            s.append("1{0:03d}".format(inject_key_code))
            s.append("0{0:03d}".format(inject_key_code))
            # we add some dummy events to slow down the keyboard entry

            # s.append("1000")
            s.append("0000")

        s = []
        for c in inject_string:
            if c == '"':
                s.append("1304105000500304")
            elif c in " ":
                inject(32)
            elif c in "\n":
                inject(13)
            elif c == '-':
                inject(45)
            elif c in "0123456789":
                code = 48 + "0123456789".index(c)
                inject(code)
            elif c in "abcdefghijklmnopqrstuvwxyz":
                code = 97 + "abcdefghijklmnopqrstuvwxyz".index(c)
                inject(code)
            else:
                raise Exception("inject_fake_input_string cannot "
                                "handle '{0}' yet".format(c))

        self.inject_fake_input_string_list(delay, s)



'''OpenGL extension VERSION.GL_1_4

This module customises the behaviour of the 
OpenGL.raw.GL.VERSION.GL_1_4 to provide a more 
Python-friendly API

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/VERSION/GL_1_4.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
from OpenGL.GL import glget
import ctypes
from OpenGL.raw.GL import _types
from OpenGL.raw.GL.VERSION.GL_1_4 import *
from OpenGL.raw.GL.VERSION.GL_1_4 import _EXTENSION_NAME

def glInitGl14VERSION():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

### END AUTOGENERATED SECTION
GL_CURRENT_FOG_COORD = GL_CURRENT_FOG_COORDINATE # alias
GL_FOG_COORD = GL_FOG_COORDINATE # alias
GL_FOG_COORD_ARRAY = GL_FOG_COORDINATE_ARRAY # alias
GL_FOG_COORD_ARRAY_POINTER = GL_FOG_COORDINATE_ARRAY_POINTER # alias
GL_FOG_COORD_ARRAY_STRIDE = GL_FOG_COORDINATE_ARRAY_STRIDE # alias
GL_FOG_COORD_ARRAY_TYPE = GL_FOG_COORDINATE_ARRAY_TYPE # alias
GL_FOG_COORD_SRC = GL_FOG_COORDINATE_SOURCE # alias



import fsui as fsui
from fsui.qt import Signal


class PagedDialog(fsui.Dialog):

    page_changed = Signal()

    def __init__(self, parent, title):
        super().__init__(parent, title)
        buttons, layout = fsui.DialogButtons.create_with_layout(self)
        buttons.create_close_button()

        hor_layout = fsui.HorizontalLayout()
        layout.add(hor_layout, fill=True, expand=True)

        layout_2 = fsui.VerticalLayout()
        # layout_2.padding_top = 20
        # layout_2.padding_bottom = 20
        # layout_2.padding_left = 20

        self.list_view = fsui.ListView(self)
        self.list_view.set_min_width(200)
        self.list_view.item_selected.connect(self.on_select_item)
        layout_2.add(self.list_view, fill=True, expand=True)
        hor_layout.add(layout_2, fill=True)

        hor_layout.add_spacer(20)

        self.page_container = fsui.Panel(self)
        self.page_container.layout = fsui.VerticalLayout()
        hor_layout.add(self.page_container, fill=True, expand=True)

        # self.layout.add_spacer(20)

        hor_layout = fsui.HorizontalLayout()
        layout.add(hor_layout, fill=True)

        self.page_titles = []
        self.pages = []

        # self.add_page(_("Hardware"), HardwarePage)
        # self.add_page(_("Hard Drives"), HardDrivesPage)
        # elf.add_page(_("Custom Options"), CustomOptionsPage)

        self.current_page = None
        self.set_size((800, 640))
        self.center_on_parent()

    # def on_close_button(self):
    #     self.end_modal(0)

    def on_select_item(self, index):
        self.set_page(index)

    def get_index(self):
        return self.list_view.get_index()

    def get_page_title(self, index):
        return self.page_titles[index][0]

    def get_page_index_by_title(self, title):
        for index, t in enumerate(self.page_titles):
            if t[0] == title:
                return index
        return None

    def add_page(self, title, function, icon=None):
        self.page_titles.append((title, icon))
        self.pages.append(function)
        self.list_view.set_items(self.page_titles)

    def set_page(self, index):
        if self.current_page:
            self.current_page.hide()
            self.page_container.layout.remove(self.current_page)
        if callable(self.pages[index]):
            page = self.pages[index](self.page_container)
            self.pages[index] = page
        else:
            page = self.pages[index]
        self.page_container.layout.add(page, fill=True, expand=True)
        self.current_page = page
        page.show()
        # print("calling self.layout.update")
        self.page_container.layout.update()
        self.layout.update()

        self.page_changed.emit()



import os
import ctypes
import getpass
import subprocess
from fsbc.path import unicode_path
from fsbc.system import windows, macosx
from fsbc.util import memoize


if windows:
    from ctypes import windll, wintypes
    _SHGetFolderPath = windll.shell32.SHGetFolderPathW
    _SHGetFolderPath.argtypes = [
        wintypes.HWND, ctypes.c_int, wintypes.HANDLE, wintypes.DWORD,
        wintypes.LPCWSTR]

    def _err_unless_zero(result):
        if result == 0:
            return result
        else:
            raise WinPathsException(
                "Failed to retrieve windows path: %s" % result)

    _SHGetFolderPath.restype = _err_unless_zero

    S_OK = 0
    CSIDL_DESKTOP = 0
    CSIDL_PERSONAL = 5
    CSIDL_APPDATA = 26
    CSIDL_MYPICTURES = 39
    CSIDL_PROFILE = 40
    CSIDL_COMMON_DOCUMENTS = 46


@memoize
def get_user_name():
    user_name = getpass.getuser()
    return user_name


@memoize
def xdg_user_dir(name):
    if windows or macosx:
        return
    try:
        process = subprocess.Popen(["xdg-user-dir", name],
                                   stdout=subprocess.PIPE)
        path = process.stdout.read().strip()
        path = path.decode("UTF-8")
        print("XDG user dir {0} => {1}".format(name, repr(path)))
    except Exception:
        path = None
    return path


class WinPathsException(Exception):
    pass


# _get_path_buf from winpaths.py
def _get_path_buf(csidl):
    path_buf = ctypes.create_unicode_buffer(wintypes.MAX_PATH)
    result = _SHGetFolderPath(0, csidl, 0, 0, path_buf)
    if result != S_OK:
        raise RuntimeError(
            "Could not find common directory for CSIDL {}".format(csidl))
    return path_buf.value


@memoize
def get_desktop_dir(allow_create=True):
    if windows:
        path = _get_path_buf(CSIDL_DESKTOP)
    else:
        path = xdg_user_dir("DESKTOP")
        if not path:
            path = os.path.join(get_home_dir(), 'Desktop')
    path = unicode_path(path)
    if allow_create and not os.path.isdir(path):
        os.makedirs(path)
    return path


@memoize
def get_documents_dir(create=False):
    if windows:
        path = _get_path_buf(CSIDL_PERSONAL)
    elif macosx:
        path = os.path.join(get_home_dir(), 'Documents')
    else:
        path = xdg_user_dir("DOCUMENTS")
        if not path:
            path = get_home_dir()
    path = unicode_path(path)
    if create and not os.path.isdir(path):
        os.makedirs(path)
    return path


@memoize
def get_common_documents_dir():
    if windows:
        path = _get_path_buf(CSIDL_COMMON_DOCUMENTS)
    else:
        raise NotImplementedError("Only for windows")
    path = unicode_path(path)
    return path


@memoize
def get_pictures_dir(allow_create=True):
    if windows:
        path = _get_path_buf(CSIDL_MYPICTURES)
    else:
        path = xdg_user_dir("PICTURES")
        if not path:
            path = os.path.join(get_home_dir(), 'Pictures')
    path = unicode_path(path)
    if allow_create and not os.path.isdir(path):
        os.makedirs(path)
    return path


@memoize
def get_home_dir():
    if windows:
        path = _get_path_buf(CSIDL_PROFILE)
    else:
        path = os.path.expanduser("~")
    path = unicode_path(path)
    return path


@memoize
def get_data_dir(allow_create=True):
    if windows:
        path = _get_path_buf(CSIDL_APPDATA)
    elif macosx:
        path = os.path.join(get_home_dir(), "Library", "Application Support")
    else:
        path = os.path.join(get_home_dir(), ".local", "share")
        path = os.environ.get("XDG_DATA_HOME", path)
    path = unicode_path(path)
    if allow_create and not os.path.exists(path):
        os.makedirs(path)
    return path



import fsui
import fsbc.system
from fsbc.util import memoize
from .Constants import Constants
from ..Settings import Settings

# LEVEL = 0xce
LEVEL = 0xeb
# LEVEL = 0xe4


class Skin(object):

    EXTRA_GROUP_MARGIN = 0

    @classmethod
    def use_unified_toolbar(cls):
        return False

    @classmethod
    def get_background_color(cls):
        c = cls._get_background_color()
        if c is None:
            return None
        return c.copy()

    @classmethod
    @memoize
    def _get_background_color(cls):
        value = Settings.get("ui_background_color")
        if len(value) == 7 and value[0] == "#":
            def convert(s):
                return int(s, 16)
            r = convert(value[1:3])
            g = convert(value[3:5])
            b = convert(value[5:7])
            return fsui.Color(r, g, b)
        if fsbc.system.windows:
            return fsui.Color(LEVEL, LEVEL, LEVEL)
        elif fsbc.system.macosx:
            return fsui.Color(237, 237, 237)
        return None

    @classmethod
    def set_background_color(cls, widget):
        color = cls.get_background_color()
        if color is not None:
            widget.set_background_color(cls.get_background_color())

    @classmethod
    def get_window_padding_left(cls):
        return 10 + cls.EXTRA_GROUP_MARGIN

    @classmethod
    def get_window_padding_right(cls):
        return 10 + cls.EXTRA_GROUP_MARGIN

    @classmethod
    def get_window_padding_bottom(cls):
        return 10 + cls.EXTRA_GROUP_MARGIN

    @classmethod
    def get_bottom_margin(cls):
        return 10 + cls.get_window_padding_bottom()

    @classmethod
    def get_bottom_panel_height(cls):
        return (Constants.SCREEN_SIZE[1] + 20 + 2 + 1 + 1 +
                cls.get_bottom_margin())



import fsui as fsui
from ...I18N import gettext
from .OptionUI import OptionUI
from fsui.extra.iconheader import NewIconHeader


class LoggingSettingsPage(fsui.Panel):

    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        self.layout = fsui.VerticalLayout()

        self.icon_header = NewIconHeader(
            self, fsui.Icon("settings", "pkg:fs_uae_workspace"),
            gettext("Logging"))
        self.layout.add(self.icon_header, fill=True, margin_bottom=20)

        label = fsui.MultiLineLabel(self, gettext(
            "The following options may affect performance, "
            "so only enable them when needed for testing or "
            "debugging purposes."), 640)
        self.layout.add(label, fill=True, margin_top=20)

        def add_option(name):
            self.layout.add(OptionUI.create_group(self, name), fill=True,
                            margin_top=10, margin_bottom=10)

        add_option("log_autoscale")
        add_option("log_input")



# FSGS - Common functionality for Fengestad Game System.
# Copyright (C) 2013  Frode Solheim <frode-code@fengestad.no>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os
import json
from fsgs.platform import PlatformHandler
from fsgs.mess.zx_spectrum import ZXSpectrumRunner
from .loader import SimpleLoader


class ZXSpectrumPlatformHandler(PlatformHandler):

    PLATFORM_NAME = "ZX Spectrum"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return ZXSpectrumLoader(fsgs)

    def get_runner(self, fsgs):
        return ZXSpectrumRunner(fsgs)


class ZXSpectrumLoader(SimpleLoader):

    def load_files(self, values):
        file_list = json.loads(values["file_list"])
        assert len(file_list) == 1

        name = file_list[0]["name"]
        sha1 = file_list[0]["sha1"]
        _, ext = os.path.splitext(name)
        ext = ext.lower()
        if ext in [".z80"]:
            key = "snapshot"
        elif ext in [".tap", ".tzx"]:
            key = "tape_drive"
        elif ext in [".dsk"]:
            key = "floppy_drive"
        else:
            return
        self.config[key] = "sha1://{0}/{1}".format(sha1, name)



from fs_uae_launcher.Signal import Signal
import fsui as fsui
from ...I18N import gettext
from ...Config import Config
from ...Options import Options
from ..HelpButton import HelpButton


class ChoiceConfigControl(fsui.Choice):

    def __init__(self, parent, option_name, choices):
        self.option_name = option_name
        # choices = [
        #     ("", gettext("Auto")),
        #     ("1", gettext("On")),
        #     ("0", gettext("Off"))]
        self.choice_values = [x[0] for x in choices]
        self.choice_labels = [x[1] for x in choices]
        super().__init__(parent, self.choice_labels)

        self.on_config(self.option_name, Config.get(self.option_name))
        Config.add_listener(self)

    def on_destroy(self):
        Config.remove_listener(self)

    def on_config(self, key, value):
        if key == self.option_name:
            print(key, self.option_name, value)
            for i, v in enumerate(self.choice_values):
                if v == value:
                    self.set_index(i)
                    break

        # FIXME: listen to implicit config also

    def on_change(self):
        index = self.get_index()
        Config.set(self.option_name, self.choice_values[index])


class BooleanConfigControl(ChoiceConfigControl):

    def __init__(self, parent, option_name):
        # choices =
        # self.choice_values = [x[0] for x in choices]
        # self.choice_labels = [x[1] for x in choices]

        super().__init__(parent, option_name, choices=[
            ("", gettext("Auto")),
            ("1", gettext("Enabled")),
            ("0", gettext("Disabled"))
        ])

        # self.on_config(self.option_name, Config.get(self.option_name))
        # Config.add_listener(self)
    #
    # def on_destroy(self):
    #     Config.remove_listener(self)
    #
    # def on_config(self, key, value):
    #     if key == self.option_name:
    #         print(key, self.option_name, value)
    #         for i, v in enumerate(self.choice_values):
    #             if v == value:
    #                 self.set_index(i)
    #                 break
    #
    #     # FIXME: listen to implicit config also
    #
    # def on_change(self):
    #     index = self.get_index()
    #     Config.set(self.option_name, self.choice_values[index])


class ConfigOptionUI(object):

    @classmethod
    def create_group(cls, parent, name):
        group = fsui.Group(parent)
        group.layout = fsui.HorizontalLayout()
        option = Options.get(name)
        group.label = fsui.Label(group, gettext(option["description"]) + ":")
        group.layout.add(group.label, margin_right=20)
        choice_values = []

        if option["type"].lower() == "boolean":
            # if option["default"] == "1":
            #     default_desc = gettext("Default ({0})").format(gettext("On"))
            # elif option["default"] == "0":
            #     default_desc = gettext("Default ({0})").format(gettext("Off"))
            # else:
            #     default_desc = gettext("Default")
            # choice_values.append(("", default_desc))
            # choice_values.append(("1", gettext("On")))
            # choice_values.append(("0", gettext("Off")))
            group.layout.add_spacer(0, expand=True)
            group.layout.add(BooleanConfigControl(group, name))

        elif option["type"].lower() == "choice":
            # for i, value in enumerate(option["values"]):
            #     if option["default"] == value[0]:
            #         default_desc = gettext("Default ({0})").format(
            #             gettext(value[1]))
            #         break
            # else:
            #     default_desc = gettext("Default")
            # choice_values.append(("", default_desc))
            # for option in option["values"]:
            #     choice_values.append((option[0], gettext(option[1])))

            group.layout.add_spacer(0, expand=True)
            choices = [("", gettext("Auto"))] + option["values"]
            group.layout.add(ChoiceConfigControl(group, name, choices))

        elif option["type"].lower() == "string":

            def on_change():
                val = text_field.get_text()
                Config.set(name, val.strip())

            text_field = fsui.TextField(group)
            # text_field.set_min_width(400)
            text_field.set_text(Config.get(name))
            text_field.on_change = on_change
            group.layout.add(text_field, expand=True)

        elif option["type"].lower() == "integer" and "min" in option \
                and "max" in option:
            current = Config.get(name)
            current_int = int(option["default"])
            if current:
                try:
                    current_int = int(current)
                except ValueError:
                    pass
            current_int = max(option["min"], min(option["max"], current_int))
            check_box = fsui.CheckBox(group, gettext("Default"))
            spin_ctrl = fsui.SpinCtrl(group, option["min"],
                                      option["max"], current_int)
            if current == "":
                check_box.check()
                spin_ctrl.disable()

            def on_checkbox():
                if check_box.is_checked():
                    spin_ctrl.set_value(int(option["default"]))
                    spin_ctrl.disable()
                    Config.set(name, "")
                else:
                    spin_ctrl.enable()

            check_box.on_change = on_checkbox

            def on_spin():
                val = spin_ctrl.get_value()
                val = max(option["min"], min(option["max"], val))
                Config.set(name, str(val))

            spin_ctrl.on_change = on_spin
            group.layout.add_spacer(0, expand=True)
            group.layout.add(check_box)
            group.layout.add(spin_ctrl, margin_left=10)

        if choice_values:

            def on_change():
                index = choice.get_index()
                Config.set(name, choice_values[index][0])

            choice_labels = [x[1] for x in choice_values]
            choice = fsui.Choice(group, choice_labels)
            current = Config.get(name)
            for i, value in enumerate(choice_values):
                if current == value[0]:
                    choice.set_index(i)
                    break
            choice.on_change = on_change
            group.layout.add_spacer(0, expand=True)
            group.layout.add(choice)
            group.widget = choice

        # group.help_button = HelpButton(
        #     parent, "http://fs-uae.net/doc/options#" + name)
        # group.layout.add(group.help_button, margin_left=10)

        return group



from fsgs.platform import PlatformHandler
from fsgs.mess.atari_7800 import Atari7800Runner
from .loader import SimpleLoader


class Atari7800PlatformHandler(PlatformHandler):

    PLATFORM_NAME = "Atari 7800"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return Atari7800Loader(fsgs)

    def get_runner(self, fsgs):
        return Atari7800Runner(fsgs)


class Atari7800Loader(SimpleLoader):
    pass



import os
import time
from urllib.request import urlopen
import threading
from fsgs.context import fsgs
from uuid import uuid4
from game_center.resources import logger
from fsgs.ogd.client import OGDClient
from fsui.qt import QImage
from game_center.glui.texturemanager import TextureManager


USE_PIL = False
USE_QT = True
# try:
#     import _no_fengestad_image
# except ImportError:
#     _no_fengestad_image = None
#     from PIL import Image
#     USE_PIL = True

error_set = set()


# class ImageData(dict):
#
#     def __init__(self):
#         dict.__init__(self)
#
#     def __del__(self):
#         #print("ImageData.__del__")
#         if "data" in self and self["data"] is not None:
#             _no_fengestad_image.free_image(self)
#
#
# def load_jpeg(path):
#     image_data = ImageData()
#     #image_data["hallo"] = "1"
#     #print image_data
#     #if draft_size is None:
#     #    draft_size = (0, 0)
#     #print(path, os.path.exists(path))
#     _no_fengestad_image.load_jpeg(image_data, path, 0, 0)
#     return image_data
#
#
# def load_png(file):
#     image_data = ImageData()
#     _no_fengestad_image.load_png(image_data, file)
#     return image_data


def get_file_for_sha1(sha1):
    # print("get_cache_path_for_sha1", sha1)
    """
    if request.args.get("is_cover", False):
        size_arg = "?size={0}".format(256)
        cache_ext = "_{0}".format(256)
    elif request.size:
        size_arg = "?w={0}&h={1}".format(request.size[0],
                request.size[1])
        cache_ext = "_{0}x{1}".format(request.size[0],
                request.size[1])
    else:
        size_arg = ""
        cache_ext = ""
    """

    # size_arg = "?size={0}".format(512)
    sha1, size_arg = sha1.split("?")
    if size_arg == "s=1x":
        cache_ext = "_1x.png"
    elif size_arg == "s=512&f=jpg":
        cache_ext = "_512.jpg"
    elif size_arg == "w=480&h=640&t=lbcover&f=jpg":
        cache_ext = "_480x640_lbcover.jpg"
    else:
        raise Exception("unrecognized size")
    
    # cache_ext = "_{0}".format(512)

    cache_dir = os.path.join(fsgs.cache_dir, "Images", sha1[:3])
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)

    cache_file = os.path.join(cache_dir, sha1 + cache_ext)
    # print(cache_file)
    if os.path.exists(cache_file):
        return cache_file

    server = OGDClient.get_server()
    cache_file_partial = cache_file + ".{0}.partial".format(str(uuid4())[:8])
    with open(cache_file_partial, "wb") as f:
        url = "http://{0}/image/{1}?{2}".format(server, sha1, size_arg)
        print(url)
        r = urlopen(url)
        # print(r)
        data = r.read()
        # h = hashlib.sha1()
        # h.update(data)
        # if h.hexdigest() == sha1:
        f.write(data)
    os.rename(cache_file_partial, cache_file)
    return cache_file


def load_image(relative_path):
    path = ""
    try:
        # print("load_image", relative_path)
        if relative_path.startswith("sha1:"):
            sha1 = relative_path[5:]
            path = get_file_for_sha1(sha1)
            # print(path)
        else:
            # base_path = os.path.join(GameCenter.get_local_data_dir())
            # path = os.path.join(base_path, relative_path)
            path = relative_path
            # if not os.path.exists(path):
            #     base_path = os.path.join(GameCenter.data_dir, "info")
            #     path = os.path.join(base_path, relative_path)

        if path in error_set:
            return None, (0, 0)
        if not os.path.exists(path):
            return None, (0, 0)

        if USE_QT:
            im = QImage(path)
            if im.format() != QImage.Format_ARGB32:
                im = im.convertToFormat(QImage.Format_ARGB32)
            bits = im.bits()
            try:
                pixels = bits.tobytes()
            except AttributeError:
                bits.setsize(im.byteCount())
                pixels = bytes(bits)
            return pixels, (im.width(), im.height())

        # if USE_PIL:
        #     # a few screenshot sizes (586 x 512, ...) has problems
        #     # when uploading to texture with GL_RGB, but works when the source is
        #     # GL_RGBA... ?
        #     im = Image.open(path)
        #     im.load()
        #     if im.mode != "RGBA":
        #         im = im.convert("RGBA")
        #     pixels = im.tostring("raw", "RGBA")
        #     #print(id(pixels))
        #     return pixels, im.size
        #
        # if path.endswith(".png"):
        #     data = load_png(path)
        # elif path.endswith(".jpg"):
        #     data = load_jpeg(path)
        # else:
        #     raise Exception("no image file extension for " + repr(path))
        # pixels = data["data"]
        # pixels = str(pixels)
        # return pixels, (data["width"], data["height"])
        # # HACK: Add data to size tuple to ensure the object lives long
        # # enough so that __del__ will not be called too soon..
        # #return pixels, (data["width"], data["height"], data)

    except Exception as e:
        print("error loading", repr(relative_path), repr(path), repr(e))
        error_set.add(path)
        return None, (0, 0)


class ImageLoader(object):

    __instance = None

    @classmethod
    def get(cls):
        if not cls.__instance:
            cls.__instance = cls()
        return cls.__instance

    def __init__(self):
        self._stop_flag = False

    def start(self):
        threading.Thread(target=self.image_loader_thread,
                         name="GameCenterImageLoaderThread").start()
        # print("FIXME: not starting ImageLoader thread")
        pass

    def image_loader_thread(self):
        logger.debug("Image loader started")
        tm = TextureManager().get()
        while not self._stop_flag:
            time.sleep(0.01)

            load_ip = None
            with tm.lock:
                for ip in tm.image_list:
                    if tm.image_dict[ip][0] is False:
                        load_ip = ip
                        break
            if load_ip:
                pixels, size = load_image(load_ip)
                tm.set_image(load_ip, pixels, size)

    def stop(self):
        print("ImageLoader.stop()")
        self._stop_flag = True



from fsgs.mednafen.mednafen import MednafenRunner


class MegaDriveRunner(MednafenRunner):

    CONTROLLER = {
        "type": "gamepad",
        "description": "Gamepad",
        "mapping_name": "megadrive",
    }

    PORTS = [
        {
            "description": "1st Controller",
            "types": [CONTROLLER]
        }, {
            "description": "2nd Controller",
            "types": [CONTROLLER]
        },
    ]

    def __init__(self, fsgs):
        super().__init__(fsgs)

    def mednafen_system_prefix(self):
        return "md"

    def mednafen_input_mapping(self, port):
        if port == 0:
            return {
                "A": "md.input.port1.gamepad.a",
                "B": "md.input.port1.gamepad.b",
                "C": "md.input.port1.gamepad.c",
                "X": "md.input.port1.gamepad.x",
                "Y": "md.input.port1.gamepad.y",
                "Z": "md.input.port1.gamepad.z",
                "UP": "md.input.port1.gamepad.up",
                "DOWN": "md.input.port1.gamepad.down",
                "LEFT": "md.input.port1.gamepad.left",
                "RIGHT": "md.input.port1.gamepad.right",
                "MODE": "md.input.port1.gamepad.mode",
                "START": "md.input.port1.gamepad.start",
            }
        elif port == 1:
            return {
                "A": "md.input.port2.gamepad.a",
                "B": "md.input.port2.gamepad.b",
                "C": "md.input.port2.gamepad.c",
                "X": "md.input.port2.gamepad.x",
                "Y": "md.input.port2.gamepad.y",
                "Z": "md.input.port2.gamepad.z",
                "UP": "md.input.port2.gamepad.up",
                "DOWN": "md.input.port2.gamepad.down",
                "LEFT": "md.input.port2.gamepad.left",
                "RIGHT": "md.input.port2.gamepad.right",
                "MODE": "md.input.port2.gamepad.mode",
                "START": "md.input.port2.gamepad.start",
            }

        # FIXME: add PAUSE button to universal gamepad config

    # def mednafen_video_size(self):
    #     return 256, 192



import os
import shutil
import subprocess
import zipfile
from fsbc.path import unicode_path, str_path
from fsbc.system import windows, macosx
from fsbc.util import chained_exception


class GameCenterUtil(object):

    @classmethod
    def get_program_path(cls, name):

        def search(search_path):
            print("search", search_path)
            if not os.path.exists(search_path):
                return None
            if macosx:
                p = os.path.join(
                    search_path, name + ".app", "Contents", "MacOS", name)
                if os.path.exists(p):
                    return p
            for item in [name] + os.listdir(search_path):
                dp = os.path.join(search_path, item)
                if item == name:
                    p = dp
                    if windows:
                        p += u".exe"
                    if os.path.exists(p):
                        return p
                if os.path.isdir(dp):
                    p = os.path.join(dp, name)
                    if windows:
                        p += u".exe"
                    print("checking", p)
                    if os.path.exists(p):
                        return p

        # try to find emulator in its own project directory
        path = search(os.path.join("..//", name, "out"))
        if path and os.path.exists(path):
            return path

        # try to find packaged emulator in a platform workspace dir

        # if fs.windows:
        #     path = search("../fs-game-center-windows/lib")
        #     if path and os.path.exists(path):
        #         return path
        # if fs.macosx:
        #     path = search("../fs-game-center-macosx/lib")
        #     if not path or not os.path.exists(path):
        #         path = search(os.path.expanduser("~/git/fs-game-center-macosx/lib"))
        #     if path and os.path.exists(path):
        #         return os.path.join(os.getcwd(), path)

        # try to find the emulator in game center's lib dir

        # path = search(fs.get_lib_dir())
        # if path and os.path.exists(path):
        #     return path

        # if not on Windows, try to find a system-installed emulator
        if not windows:
            path = "/usr/bin/" + name
            if os.path.exists(path):
                print("checking", path)
                return path
            path = "/usr/local/bin/" + name
            if os.path.exists(path):
                print("checking", path)
                return path

        # or on Mac, search Applications dir for bundle
        if macosx:
            path = search("/Applications")
            if path and os.path.exists(path):
                return path

    @classmethod
    def run_program(cls, name, **kwargs):
        path = cls.get_program_path(name)
        if not path:
            raise Exception("Could not find program {0}".format(name))
        print("program", name, "at", path)
        path = os.path.join(os.getcwd(), path)
        path = os.path.normpath(path)
        print("program path (normalized):", path)
        dir_name = os.path.dirname(path)
        kwargs["env"]["LD_LIBRARY_PATH"] = str_path(dir_name)
        kwargs["args"].insert(0, path)
        print("run:", kwargs["args"])
        print("cwd:", kwargs.get("cwd", ""))
        return subprocess.Popen(**kwargs)
        # for entry_point in pkg_resources.iter_entry_points(
        #         "fengestad.program", name):
        #     # return first available, load should return a callable
        #     run_func = entry_point.load()
        #     break
        # else:
        #     raise Exception("Could not find program {0}".format(name))
        # print("run program ", name, "with args", kwargs["args"])
        # #print("run program ", name, "with args", kwargs)
        # return run_func(**kwargs)

    @classmethod
    def copy_folder_tree(cls, source_path, dest_path, overwrite=False):
        for item in os.listdir(source_path):
            if item[0] == ".":
                continue
            item_path = os.path.join(source_path, item)
            dest_item_path = os.path.join(dest_path, item)
            if os.path.isdir(item_path):
                if not os.path.exists(dest_item_path):
                    os.makedirs(dest_item_path)
                cls.copy_folder_tree(item_path, dest_item_path)
            else:
                if overwrite or not os.path.exists(dest_item_path):
                    shutil.copy(item_path, dest_item_path)

    @classmethod
    def unpack(cls, archive, destination):
        if archive.endswith(".7z"):
            cls.unpack_7zip(archive, destination)
        elif archive.endswith(".zip"):
            cls.unpack_zip(archive, destination)
        else:
            raise Exception("do not know how to unpack " + repr(archive))

    @classmethod
    def unpack_zip(cls, archive, destination):
        try:
            print("unpack", archive, "to", destination)
            zip = zipfile.ZipFile(archive, "r")

            def extract_members(z):
                for name in z.namelist():
                    if ".." in name:
                        continue
                    if not os.path.normpath(os.path.join(
                            destination, name)).startswith(destination):
                        raise Exception("Invalid path")
                    yield name

            zip.extractall(path=destination, members=extract_members(zip))
            zip.close()
        except Exception as e:
            print(e)
            raise chained_exception(Exception("Could not unpack game"), e)

    @classmethod
    def unpack_7zip(cls, archive, destination):
        try:
            print("unpack", archive, "to", destination)
            if not os.path.exists(destination):
                os.makedirs(destination)
            # plugin = pyapp.plug.get_plugin("no.fengestad.7zip")
            args = ["x", "-o" + destination, archive]
            process = cls.run_program("7za", args=args,
                                      stdout=subprocess.PIPE)
            # process = plugin.sevenz(args, stdout=subprocess.PIPE)
            line = process.stdout.readline()
            while line:
                print(line[:-1])
                # if fs.windows:
                line = unicode_path(line[:-len(os.linesep)])
                if line.startswith("Extracting  "):
                    parts = line.rsplit(os.sep, 1)
                    if len(parts) == 2:
                        # text = "Extracting " + parts[1]
                        # self.on_status(status="Extracting", sub_status=parts[1])
                        pass
                line = process.stdout.readline()
            return_code = process.wait()
            if return_code != 0:
                raise Exception("7z return non-zero")
        except Exception as e:
            print(e)
            raise chained_exception(Exception("Could not unpack game"), e)



import fsui
from fs_uae_launcher.I18N import gettext
from ..behaviors.configbehavior import ConfigBehavior


class OverrideWarning(fsui.ImageView):

    def __init__(self, parent, name):
        fsui.ImageView.__init__(self, parent,fsui.Image(
            "fs_uae_launcher:res/16/warning_2.png"))
        setattr(self, "on_{0}_config".format(name), self.on_config)
        ConfigBehavior(self, [name])
        text = gettext(
            "Option {name} is overridden by current configuration".format(
                name=name))
        self.set_tool_tip(text)

    def on_config(self, value):
        self.show_or_hide(bool(value))



from .mess import MESSRunner


class ZXSpectrumRunner(MESSRunner):

    JOYSTICK = {
        "type": "joystick",
        "description": "Joystick",
        "mapping_name": "zx-spectrum-joystick",
    }

    PORTS = [
        {
            "description": "Joystick Port",
            "types": [JOYSTICK]
        },
    ]

    def mess_configure(self):
        # FIXME: hack
        # self.mess_configure_floppies(["cassette"])

        if self.config["snapshot"]:
            # snapshot loads automatically, no need to do anything special
            self.add_arg("-{0}".format("snapshot"),
                         self.get_game_file("snapshot"))
        elif self.config["floppy_drive"]:
            # pressing enter to select "loader" in the menu
            self.inject_fake_input_string(160, "{0}\n".format(""))
            self.add_arg("-{0}".format("floppydisk1"),
                         self.get_game_file("floppy_drive"))
        elif self.config["tape_drive"]:
            self.inject_fake_input_string_list(160, [
                "1106", "0106", "0000",  # j (LOAD)
                "1032", "0032", "0000",  # space
                "1303", "0000",          # right shift (press)
                "1112", "0112", "0000",  # p (with symbol shift = ")
                "0303", "0000",          # right shift (release)
                "1303", "0000",          # right shift (press)
                "1112", "0112", "0000",  # p (with symbol shift = ")
                "0303", "0000",          # right shift (release)
                "1013", "0013", "0000",  # return
            ])
            self.add_arg("-{0}".format("cassette"),
                         self.get_game_file("tape_drive"))

    def mess_full_keyboard(self):
        return True

    def mess_input_mapping(self, port):
        return {
            "UP": 'tag=":KEMPSTON" type="KEYBOARD" mask="8" defvalue="0"',
            "DOWN": 'tag=":KEMPSTON" type="KEYBOARD" mask="4" defvalue="0"',
            "LEFT": 'tag=":KEMPSTON" type="KEYBOARD" mask="2" defvalue="0"',
            "RIGHT": 'tag=":KEMPSTON" type="KEYBOARD" mask="1" defvalue="0"',
            "1": 'tag=":KEMPSTON" type="KEYBOARD" mask="16" defvalue="0"',
        }

    def get_game_refresh_rate(self):
        # Refresh rate values retrieved from MESS
        if self.config["model"] == "spectrum128":
            return 50.021084
        else:
            return 50.080128

    def mess_romset(self):
        # MESS system names:
        # sp3e8bit          ZX Spectrum +3e 8bit IDE
        # sp3eata           ZX Spectrum +3e 8bit ZXATASP
        # sp3ezcf           ZX Spectrum +3e 8bit ZXCF
        # spec128           ZX Spectrum 128
        # spec80k           ZX Spectrum 80K
        # specide           ZX Spectrum IDE
        # specpl2a          ZX Spectrum +2a
        # specpl3e          ZX Spectrum +3e
        # specpls2          ZX Spectrum +2
        # specpls3          ZX Spectrum +3
        # spectrum          ZX Spectrum
        if self.config["model"] == "spectrum128":
            return "spec128", SPEC128_ROMS
        if self.config["model"] == "spectrum+3":
            return "specpls3", SPECPLS3_ROMS
        else:
            return "spectrum", SPECTRUM_ROMS


SPECTRUM_ROMS = [
    ("f9d23f25640c51bcaa63e21ed5dd66bb2d5f63d4", "1986es.rom"),
    ("9e535e2e24231ccb65e33d107f6d0ceb23e99477", "48e.rom"),
    ("e62a431b0938af414b7ab8b1349a18b3c4407f70", "48turbo.rom"),
    ("ab3c36daad4325c1d3b907b6dc9a14af483d14ec", "bsrom118.rom"),
    ("2ee2dbe6ab96b60d7af1d6cb763b299374c21776", "bsrom140.rom"),
    ("795c20324311dd5a56300e6e4ec49b0a694ac0b3", "deutsch.rom"),
    ("51165cde68e218512d3145467074bc7e786bf307", "groot.rom"),
    ("a701c3d4b698f7d2be537dc6f79e06e4dbc95929", "gw03.rom"),
    ("2a9745ba3b369a84c4913c98ede66ec87cb8aec1", "hdt-iso.rom"),
    ("dee814271c4d51de257d88128acdb324fb1d1d0d", "imc.rom"),
    ("5752e6f789769475711b91e0a75911fa5232c767", "iso8bm.rom"),
    ("04adbdb1380d6ccd4ab26ddd61b9ccbba462a60f", "isomoje.rom"),
    ("d7f02ed66455f1c08ac0c864c7038a92a88ba94a", "jgh.rom"),
    ("c103e89ef58e6ade0c01cea0247b332623bd9a30", "plus4.rom"),
    ("0853e25857d51dd41b20a6dbc8e80f028c5befaa", "psycho.rom"),
    ("c58ff44a28db47400f09ed362ca0527591218136", "sc01.rom"),
    ("5ea7c2b824672e914525d1d5c419d71b84a426a2", "spectrum.rom"),
    ("84ea64af06adaf05e68abe1d69454b4fc6888505", "turbo2_3.rom"),
    ("21ad93ffe41a4458704c866cca2754f066f6a560", "turbo4_4.rom"),
]


SPEC128_ROMS = [
    ("4f4b11ec22326280bdb96e3baf9db4b4cb1d02c5", "zx128_0.rom"),
    ("80080644289ed93d71a1103992a154cc9802b2fa", "zx128_1.rom"),
    ("968937b1c750f0ef6205f01c6db4148da4cca4e3", "zx128s0.rom"),
    ("bea3f397cc705eafee995ea629f4a82550562f90", "zx128s1.rom"),
]


SPECPLS3_ROMS = [
    ("4e5d114b72d464cefdde0566457f52a3c0c1cae2", "p3_01_4m.rom"),
    ("4e5d114b72d464cefdde0566457f52a3c0c1cae2", "p3_01_cm.rom"),
    ("752cdd6a083ab9910348995e483541d60bb6372b", "p3_23_4m.rom"),
    ("d062765ceb1f3cd2c94ea51cb737cac7ad6151b4", "p3_23_cm.rom"),
    ("e319ed08b4d53a5e421a75ea00ea02039ba6555b", "pl3-0.rom"),
    ("c9969fc36095a59787554026a9adc3b87678c794", "pl3-1.rom"),
    ("22e50c6ba4157a3f6a821bd9937cd26e292775c6", "pl3-2.rom"),
    ("65f031caa8148a5493afe42c41f4929deab26b4e", "pl3-3.rom"),
    ("500c0945760abeefcbd08bc22c0d07b14b336cf0", "plus341.rom"),
    ("e9b0a60a1a8def511d59090b945d175bdc646346", "plus3sp0.rom"),
    ("4e48f196427596c7990c175d135c15a039c274a4", "plus3sp1.rom"),
    ("09fc005625589ef5992515957ce7a3167dec24b2", "plus3sp2.rom"),
    ("ec8f644a81e2e9bcb58ace974103ea960361bad2", "plus3sp3.rom"),
]



'''OpenGL extension ARB.explicit_attrib_location

This module customises the behaviour of the 
OpenGL.raw.GL.ARB.explicit_attrib_location to provide a more 
Python-friendly API

Overview (from the spec)
	
	This extension provides a method to pre-assign attribute locations to
	named vertex shader inputs and color numbers to named fragment shader
	outputs.  This allows applications to globally assign a particular
	semantic meaning, such as diffuse color or vertex normal, to a particular
	attribute location without knowing how that attribute will be named in any
	particular shader.

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/ARB/explicit_attrib_location.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
from OpenGL.GL import glget
import ctypes
from OpenGL.raw.GL import _types
from OpenGL.raw.GL.ARB.explicit_attrib_location import *
from OpenGL.raw.GL.ARB.explicit_attrib_location import _EXTENSION_NAME

def glInitExplicitAttribLocationARB():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

### END AUTOGENERATED SECTION


# noinspection PyUnresolvedReferences
from fsbc.user import get_home_dir
# noinspection PyUnresolvedReferences
from fsbc.util import memoize
from fsbc.Paths import Paths


def expand_path(path):
    return Paths.expand_path(path)


def get_real_case(path):
    return Paths.get_real_case(path)



import fsui
from fs_uae_launcher.I18N import gettext
from fs_uae_launcher.Options import Option
from fs_uae_launcher.ui.config.configpanel import ConfigPanel
from fs_uae_launcher.ui.config.ExpansionsGroup import ExpansionsGroup


class ExpansionsPanel(ConfigPanel):

    def __init__(self, parent):
        ConfigPanel.__init__(self, parent)

        # self.expansions_group = ExpansionsGroup(self)
        # self.layout.add(self.expansions_group, fill=True)

        heading_label = fsui.HeadingLabel(self, gettext("Expansions"))
        self.layout.add(heading_label, margin=10)
        self.layout.add_spacer(0)

        # self.add_option(Option.CPU)
        self.add_option(Option.ACCELERATOR)
        self.add_option(Option.ACCELERATOR_MEMORY)
        self.add_option(Option.BLIZZARD_SCSI_KIT)

        self.add_option(Option.GRAPHICS_CARD)
        self.add_option(Option.GRAPHICS_MEMORY)

        self.add_option(Option.SOUND_CARD)
        self.add_option(Option.BSDSOCKET_LIBRARY)



from xml.etree.cElementTree import ElementTree
import fsui as fsui
from ..I18N import gettext
from .GameListView import GameListView


class PickGameDialog(fsui.LegacyDialog):

    def __init__(self, parent):
        fsui.LegacyDialog.__init__(self, parent, "Choose Game")
        self.info = {}
        self.layout = fsui.VerticalLayout(padding=20)
        self.text = fsui.TextField(self)
        self.text.on_change = self.on_text
        self.layout.add(self.text, fill=True)
        self.layout.add_spacer(20)
        self.view = GameListView(self)
        self.view.set_game_info = self.set_game_info
        self.view.set_min_width(600)
        self.view.set_min_height(300)
        self.layout.add(self.view, expand=True, fill=True)
        self.layout.add_spacer(20)

        self.link_field = fsui.TextField(self)
        self.layout.add(self.link_field, fill=True)
        self.layout.add_spacer(20)

        hlayout = fsui.HorizontalLayout()
        self.layout.add(hlayout, fill=True)
        hlayout.add_spacer(0, expand=True)
        self.cancel_button = fsui.Button(self, gettext("Cancel"))
        self.cancel_button.activated.connect(self.on_cancel_button)
        hlayout.add(self.cancel_button)
        hlayout.add_spacer(10)
        self.ok_button = fsui.Button(self, gettext("OK"))
        self.ok_button.activated.connect(self.on_ok_button)
        hlayout.add(self.ok_button)

        self.text.focus()
        self.set_size(self.layout.get_min_size())
        self.center_on_parent()

    def set_game_info(self, info):
        self.info = info
        tree = ElementTree()
        tree.parse(info["path"])
        link_node = tree.find("link")
        if link_node is not None:
            self.link_field.set_text(link_node.text.strip())
        else:
            self.link_field.set_text("")

    def on_text(self):
        self.view.search(self.text.get_text())

    def get_game_info(self):
        return self.info

    def get_game_uuid(self):
        return self.info.get("uuid")

    def on_ok_button(self):
        self.end_modal(bool(self.info.get("uuid")))

    def on_cancel_button(self):
        self.end_modal(False)



from fsbc.Application import app
import fsui as fsui
from ...I18N import gettext
from ...Settings import Settings


class FullscreenToggleButton(fsui.ImageButton):

    def __init__(self, parent):
        self.windowed_icon = fsui.Image(
            "fs_uae_launcher:res/windowed_16.png")
        self.fullscreen_icon = fsui.Image(
            "fs_uae_launcher:res/fullscreen_16.png")
        fsui.ImageButton.__init__(self, parent, self.windowed_icon)
        self.set_tooltip(
            gettext("Toggle Between Windowed and Full-Screen Mode"))
        self.set_min_width(40)
        self.fullscreen_mode = False
        self.on_setting("fullscreen", app.settings["fullscreen"])
        Settings.add_listener(self)

    def on_destroy(self):
        Settings.remove_listener(self)

    def on_setting(self, key, value):
        if key == "fullscreen":
            if value == "1":
                self.fullscreen_mode = True
                self.set_image(self.fullscreen_icon)
            else:
                self.fullscreen_mode = False
                self.set_image(self.windowed_icon)

    def on_activate(self):
        if self.fullscreen_mode:
            app.settings["fullscreen"] = "0"
        else:
            app.settings["fullscreen"] = "1"



from fsgs.platform import PlatformHandler
from fsgs.dosbox.dos import DOSRunner
from .loader import SimpleLoader


class DOSPlatformHandler(PlatformHandler):

    PLATFORM_NAME = "DOS"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return DOSLoader(fsgs)

    def get_runner(self, fsgs):
        return DOSRunner(fsgs)


class DOSLoader(SimpleLoader):

    def load_files(self, values):
        self.config["file_list"] = values["file_list"]

    def load_extra(self, values):
        self.config["hd_startup"] = values["hd_startup"]



from fsgs.ogd.client import OGDClient
import fsui
from fsbc.Application import app
from fs_uae_workspace.shell import SimpleApplication
from fs_uae_launcher.res import gettext


class LogoutWindow(fsui.Dialog):

    def __init__(self):
        super().__init__(None, gettext("Log Out from Your OAGD.net Account"))
        self.set_icon(fsui.Icon("password", "pkg:fs_uae_workspace"))

        self.layout = fsui.VerticalLayout()
        self.layout.set_padding(20, 20, 20, 20)

        heading_layout = fsui.HorizontalLayout()
        self.layout.add(heading_layout)
        heading_layout.add(fsui.ImageView(
            self, fsui.Image("fs_uae_workspace:res/48/password.png")))
        heading_layout.add_spacer(20)
        heading_layout_2 = fsui.VerticalLayout()
        heading_layout.add(
            heading_layout_2, expand=True, fill=False, valign=0.5)
        heading_layout_2.add(fsui.HeadingLabel(
            self, gettext("Log Out from Your OAGD.net Account")))
        heading_layout_2.add_spacer(2)
        heading_layout_2.add(fsui.Label(
            self, gettext("While logged out you will not get "
                          "database updates")))

        self.layout.add_spacer(20)
        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        self.created_label = fsui.Label(self, "")
        hori_layout.add(self.created_label, expand=True)
        hori_layout.add_spacer(20)

        self.logout_button = fsui.Button(self, gettext("Log Out"))
        # self.logout_button.disable()
        self.logout_button.activated.connect(self.on_logout_activated)
        hori_layout.add(self.logout_button)

        self.close_button = fsui.Button(self, gettext("Close"))
        self.close_button.activated.connect(self.on_close_activated)
        hori_layout.add(self.close_button, margin_left=10)

        self.set_size(self.layout.get_min_size())
        self.center_on_parent()

    def __del__(self):
        print("LogoutWindow.__del__")

    def on_close_activated(self):
        self.close()

    def on_logout_activated(self):
        auth_token = app.settings["database_auth"]
        if auth_token:
            task = OGDClient().logout_task(auth_token)
            # task.progressed.connect(self.progress)
            task.succeeded.connect(self.close)
            # task.failed.connect(fsui.error_function(gettext("Login Failed")))
            task.failed.connect(self.on_failure)
            task.start()
        else:
            # this is not a normal case, no auth token stored, but clear
            # all auth-related settings just in case
            app.settings["database_auth"] = ""
            app.settings["database_username"] = ""
            # app.settings["database_email"] = ""
            app.settings["database_password"] = ""
            self.on_close()

    def on_failure(self, message):
        fsui.show_error(message, parent=self.get_window())

    # def on_progress(self):
    #    print("on_progress")

    # def on_success(self):
    #     app.settings["database_auth"] = ""
    #     app.settings["database_username"] = ""
    #     app.settings["database_email"] = ""
    #     app.settings["database_password"] = ""
    #     self.close()

application = SimpleApplication(LogoutWindow)



import os
import sys
import traceback
from fsbc.Application import app, Application
import fsbc.fs as fs
from fsbc.Paths import Paths
from fsbc.user import get_home_dir, get_documents_dir
from fsbc.user import get_common_documents_dir
from fsbc.util import memoize
from fsbc.system import windows


class FSGSDirectories(object):

    base_dir = ""
    initialized = False

    @classmethod
    def initialize(cls):
        if cls.initialized:
            return
        cls.initialized = True
        for arg in sys.argv[1:]:
            if arg.startswith("--"):
                if "=" in arg:
                    key, value = arg[2:].split("=", 1)
                    key = key.replace("-", "_")
                    if key == "base_dir":
                        value = os.path.abspath(value)
                        print("setting base_dir")
                        cls.base_dir = value
        if cls.base_dir:
            print(
                "base_dir was specified so we will not check for "
                "portable dir")
        else:
            cls.setup_portable()
        if not cls.base_dir:
            print("base_dir not decided yet, checking FS_UAE_BASE_DIR")
            if "FS_UAE_BASE_DIR" in os.environ and \
                    os.environ["FS_UAE_BASE_DIR"]:
                print("base dir specified by FS_UAE_BASE_DIR")
                cls.base_dir = os.environ["FS_UAE_BASE_DIR"]
        if cls.base_dir:
            print("base dir is", cls.base_dir)
            try:
                cls.base_dir = Paths.get_real_case(cls.base_dir)
            except:
                traceback.print_exc()
                print("WARNING: error looking up real case for base dir")
            print("base dir (normalized) is", cls.base_dir)
        else:
            print("using default base_dir")

    @classmethod
    def setup_portable(cls):
        path = Application.executable_dir()
        # path = os.path.dirname(os.path.abspath(sys.executable))
        last = ""
        while not last == path:
            portable_ini_path = os.path.join(path, "Portable.ini")
            print("checking", portable_ini_path)
            if os.path.exists(portable_ini_path):
                print("detected portable dir", path)
                cls.base_dir = path
                return
            last = path
            path = os.path.dirname(path)
        print("no Portable.ini found in search path")

    @classmethod
    def read_custom_path(cls, name):
        for app_name in ["fs-uae-launcher", "fs-uae"]:
            key_path = os.path.join(fs.get_app_config_dir(app_name), name)
            print("- checking", key_path)
            if os.path.exists(key_path):
                try:
                    with open(key_path, "r", encoding="UTF-8") as f:
                        path = f.read().strip()
                        break
                except Exception as e:
                    print("error reading custom path", repr(e))
        else:
            return None
        path_lower = path.lower()
        if path_lower.startswith("$home/") or path_lower.startswith("$home\\"):
            path = os.path.join(get_home_dir(), path[6:])
        return path

    @classmethod
    def get_base_dir(cls):
        cls.initialize()
        path = cls.base_dir
        if path:
            return path
        return cls._get_base_dir()

    @classmethod
    @memoize
    def _get_base_dir(cls):
        path = cls.read_custom_path("base-dir")
        if not path:
            path = os.path.join(get_documents_dir(True), "FS-UAE")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        print("base dir is", path)
        return path

    @classmethod
    def get_default_search_path(cls):
        paths = []
        path = cls.get_base_dir()
        # if windows:
        #     path = path.replace("/", "\\")
        paths.append(path)

        if windows:
            path = get_common_documents_dir()
            path = os.path.join(path, "Amiga Files")
            if os.path.exists(path):
                paths.append(path)
        return paths

    @classmethod
    def get_whdload_dir(cls):
        path = os.path.join(cls.get_hard_drives_dir(), "WHDLoad")
        if os.path.exists(path):
            path = Paths.get_real_case(path)
            return path
        return None

    @classmethod
    def get_configurations_dir(cls):
        path = os.path.join(cls.get_base_dir(), "Configurations")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    def get_controllers_dir(cls):
        path = os.path.join(cls.get_base_dir(), "Controllers")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    def get_kickstarts_dir(cls):
        path = os.path.join(cls.get_base_dir(), "Kickstarts")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    def get_downloads_dir(cls):
        path = os.path.join(cls.get_base_dir(), "Downloads")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    def get_floppies_dir(cls):
        path = os.path.join(cls.get_base_dir(), "Floppies")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    def get_hard_drives_dir(cls):
        path = os.path.join(cls.get_base_dir(), "Hard Drives")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    def get_cdroms_dir(cls):
        path = os.path.join(cls.get_base_dir(), "CD-ROMs")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    def get_logs_dir(cls):
        path = os.path.join(cls.get_cache_dir(), "Logs")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    def get_data_dir(cls):
        path = os.path.join(cls.get_base_dir(), "Data")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    def get_launcher_dir(cls):
        return cls.get_data_dir()

    @classmethod
    def get_titles_dir(cls):
        path = os.path.join(cls.get_base_dir(), "Titles")
        return path

    @classmethod
    def get_save_states_dir(cls):
        path = os.path.join(cls.get_base_dir(), "Save States")
        if not os.path.exists(path):
            os.makedirs(path)
        path = Paths.get_real_case(path)
        return path

    @classmethod
    def screenshots_output_dir(cls):
        path = app.settings["screenshots_output_dir"]
        if not path:
            path = os.path.join(get_documents_dir(), "Screenshots")
        if not os.path.exists(path):
            os.makedirs(path)
        return path

    @classmethod
    def get_screenshots_dir(cls):
        path = os.path.join(cls.get_base_dir(), "Screenshots")
        return path

    @classmethod
    def get_images_dir(cls):
        path = cls.get_base_dir()
        return path

    @classmethod
    def get_covers_dir(cls):
        path = os.path.join(cls.get_base_dir(), "Covers")
        return path

    @classmethod
    def get_themes_dir(cls):
        path = os.path.join(cls.get_base_dir(), "Themes")
        if not os.path.exists(path):
            os.makedirs(path)
        return path

    @classmethod
    def get_plugins_dir(cls):
        path = os.path.join(cls.get_base_dir(), "Plugins")
        if not os.path.exists(path):
            return None
        return path

    @classmethod
    def get_cache_dir(cls):
        path = os.path.join(cls.get_base_dir(), "Cache")
        if not os.path.exists(path):
            os.makedirs(path)
        if not os.path.exists(os.path.join(path, ".nobackup")):
            open(os.path.join(path, ".nobackup"), "w").close()
        return path

    @classmethod
    def get_files_dirs(cls):
        paths = [
            cls.get_floppies_dir(),
            cls.get_cdroms_dir(),
        ]
        return paths

    @classmethod
    def get_titles_dirs(cls):
        paths = [
            cls.get_titles_dir(),
        ]
        return paths

    @classmethod
    def get_screenshots_dirs(cls):
        paths = [
            cls.get_screenshots_dir(),
        ]
        return paths

    @classmethod
    def get_images_dirs(cls):
        paths = [
            cls.get_images_dir(),
        ]
        return paths

    @classmethod
    def get_covers_dirs(cls):
        paths = [
            cls.get_covers_dir(),
        ]
        return paths

    @classmethod
    def get_themes_dirs(cls):
        paths = [cls.get_themes_dir()]
        return paths



'''OpenGL extension ARB.shading_language_include

This module customises the behaviour of the 
OpenGL.raw.GL.ARB.shading_language_include to provide a more 
Python-friendly API

Overview (from the spec)
	
	This extension introduces a #include GLSL directive to allow reusing
	the same shader text in multiple shaders and defines the semantics
	and syntax of the names allowed in #include directives. It also
	defines API mechanisms to define the named string backing a
	#include.

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/ARB/shading_language_include.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
from OpenGL.GL import glget
import ctypes
from OpenGL.raw.GL import _types
from OpenGL.raw.GL.ARB.shading_language_include import *
from OpenGL.raw.GL.ARB.shading_language_include import _EXTENSION_NAME

def glInitShadingLanguageIncludeARB():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

### END AUTOGENERATED SECTION


import sys
from fsbc.Application import app
from fsui.qt import Qt, QMainWindow, QGLWidget, QKeyEvent
from game_center.glui.input import InputHandler
from .event import Event


# noinspection PyPep8Naming
class GLWidget(QGLWidget):

    def __init__(self, parent, callbacks):
        QGLWidget.__init__(self, parent)
        set_black_background(self)
        self._callbacks = callbacks
        self._initialized = False

        palette = self.palette()
        palette.setColor(self.backgroundRole(), Qt.blue)
        self.setPalette(palette)
        self.setAutoFillBackground(True)
        self.setStyleSheet("background-color: black;")

    def initializeGL(self):
        print("GLWidget.initializeGL")
        # moved initialization to resizeGL due to issues on OS X

    def resizeGL(self, width, height):
        print("GLWidget.resizeGL", width, height)
        if not self._initialized:
            if width == 160 and height == 160:
                # work around bug(?) on os x
                return
            self._callbacks.initialize()
            self._initialized = True
        self._callbacks.resize(width, height)

    def paintGL(self):
        self._callbacks.render()


# noinspection PyPep8Naming
class QtWindow(QMainWindow):

    def __init__(self, callback, interval):
        # QMainWindow.__init__(self, None, Qt.X11BypassWindowManagerHint)
        QMainWindow.__init__(self)
        set_black_background(self)
        if app.name == "fs-uae-arcade":
            self.setWindowTitle("FS-UAE Arcade")
        else:
            self.setWindowTitle("FS Game Center")
        self.gl_widget = GLWidget(self, callback)
        if "--show-cursor" not in sys.argv:
            self.setCursor(Qt.BlankCursor)
            self.gl_widget.setCursor(Qt.BlankCursor)
        self.setCentralWidget(self.gl_widget)
        self.startTimer(interval)
        self.callback = callback
        # self.resize(940, 540)

    def restore_window_if_necessary(self):
        pass
        # print("restore_window_if_necessary")
        # print("is minimized:", self.isMinimized())
        # # under Gnome 3, the window is minized when launching FS-UAE
        # # full-screen from full-screen arcade/game center.
        # if self.isMinimized():
        #     # self.showFullScreen()
        #     self.setWindowState(self.windowState() ^ Qt.WindowMinimized)
        #     self.activateWindow()

    def timerEvent(self, event):
        self.callback.timer()
        self.gl_widget.updateGL()
        # self.callback()

    def keyPressEvent(self, event):
        assert isinstance(event, QKeyEvent)
        # print(event.isAutoRepeat(), event.type())
        if event.isAutoRepeat():
            return
        InputHandler.add_event(Event.create_key_event(event))
        text = event.text()
        if text and text in TEXT_WHITE_LIST:
            # We don't want special characters such as return, backspace
            # and escape (etc) to be sent as text events. For now, we use
            # a simple white list.
            InputHandler.add_event({
                "type": "text",
                "text": event.text(),
            })

    def keyReleaseEvent(self, event):
        assert isinstance(event, QKeyEvent)
        # print(QKeyEvent, event.isAutoRepeat(), event.type())
        if event.isAutoRepeat():
            return
        InputHandler.add_event(Event.create_key_event(event))


TEXT_WHITE_LIST = (
    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.,- ")


def set_black_background(widget):
    palette = widget.palette()
    palette.setColor(widget.backgroundRole(), Qt.blue)
    widget.setPalette(palette)
    widget.setAutoFillBackground(True)
    widget.setStyleSheet("background-color: black;")



import os
import sys
import io
import subprocess
from collections import defaultdict
from fsbc.Application import app
from fsbc.system import windows, macosx
from fsbc.task import current_task
from fsgs.refreshratetool import RefreshRateTool
from fsgs.FSGSDirectories import FSGSDirectories
from fsgs.util.gamenameutil import GameNameUtil


class Port(object):

    def __init__(self, name):
        self.name = name
        self.types = []
        self.index = 0
        self.device = None

        # FIXME: remove
        self.device_id = None
        self.device_config = None

    @property
    def type(self):
        return self.types[self.index]["type"]

    @property
    def mapping_name(self):
        return self.types[self.index]["mapping_name"]

    @property
    def description(self):
        return self.types[self.index]["description"]


class GameRunner(object):

    PORTS = []

    def __init__(self, fsgs):
        self.fsgs = fsgs
        self.__env = {}
        self.__args = []

        self.config = defaultdict(str)
        for key, value in app.settings.values.items():
            # FIXME: re-enable this check?
            # if key in Config.config_keys:
            #     print("... ignoring config key from settings:", key)
            #     continue
            self.config[key] = value

        for key, value in self.fsgs.config.items():
            self.config[key] = value

        self.ports = []
        for port_info in self.PORTS:
            port = Port(port_info["description"])
            port.types = port_info["types"]
            self.ports.append(port)

        self.__vsync = False
        self.__game_temp_file = None

        # default current working directory for emulator
        self.cwd = self.create_temp_dir("cwd")
        self.home = self.cwd

    # self.inputs.append(self.create_input(
    #         name='Controller {0}'.format(i + 1),
    #         type='megadrive',
    #         description='Gamepad'))

    def use_fullscreen(self):
        # FIXME: not a very nice hack to hard-code application name here...
        if app.name == "fs-uae-arcade":
            return True
        if app.settings["fullscreen"] == "0":
            return False
        return True

    def use_vsync(self):
        # if "--no-vsync" in sys.argv:
        #     return False
        # return True
        return self.config.get("video_sync") == "1"

    def use_doubling(self):
        return True

    def use_smoothing(self):
        return True

    def use_stretching(self):
        if "--no-stretch" in sys.argv:
            return False
        if app.settings["stretch"] == "0":
            return False
        return True

    def use_audio_frequency(self):
        if app.settings["audio_frequency"]:
            try:
                return int(app.settings["audio_frequency"])
            except ValueError:
                pass
        return 48000

    def screenshots_dir(self):
        return FSGSDirectories.screenshots_output_dir()

    def screenshots_name(self):
        return GameNameUtil.create_fs_name(self.get_name())

    # def get_game_name(self):
    #     return self.config["game_name"]
    #
    # def get_variant_name(self):
    #     return self.config["variant_name"]

    def get_platform_name(self):
        p = self.config["platform"].lower()
        print(p)
        if p == "atari-7800":
            return "Atari 7800"
        if p == "amiga":
            return "Amiga"
        if p == "cdtv":
            return "CDTV"
        if p == "cd32":
            return "CD32"
        raise Exception("Unrecognized platform")

    def screen_size(self):
        refresh_rate_tool = RefreshRateTool()
        # FIXME: screen size monitor size
        width = refresh_rate_tool.get_current_mode()["width"]
        height = refresh_rate_tool.get_current_mode()["height"]
        # if width > 2 * height:
        #     print("width > 2 * height, assuming dual-monitor setup...")
        #     return width // 2, height
        return width, height

    def screen_rect(self):
        try:
            desktop = app.qapplication.desktop()
        except AttributeError:
            # no QApplication, probably not running via QT
            # FIXME: log warning
            return 0, 0, 640, 480
        screens = []
        for i in range(desktop.screenCount()):
            geometry = desktop.screenGeometry(i)
            screens.append([geometry.x(), i, geometry])
        screens.sort()
        monitor = self.config.get("monitor", "")
        if monitor == "left":
            mon = 0
        elif monitor == "middle-right":
            mon = 2
        elif monitor == "right":
            mon = 3
        else:  # middle-left
            mon = 1
        display = round(mon / 3 * (len(screens) - 1))
        geometry = screens[display][2]
        return geometry.x(), geometry.y(), geometry.width(), geometry.height()

    def get_screen_width(self):
        """deprecated"""
        return self.screen_size()[0]

    def get_screen_height(self):
        """deprecated"""
        return self.screen_size()[1]

    def get_name(self):
        # return "{0} ({1}, {2})".format(
        #     self.get_game_name(), self.get_platform_name(),
        #     self.get_variant_name())
        return "{0} ({1}, {2})".format(
            self.fsgs.game.name, self.fsgs.game.platform.name,
            self.fsgs.game.variant.name)

    def get_state_dir(self):
        state_dir = os.path.join(
            FSGSDirectories.get_save_states_dir(),
            GameNameUtil.create_fs_name(self.get_name()))
        if not os.path.exists(state_dir):
            os.makedirs(state_dir)
        if self.fsgs.game.variant.uuid:
            uuid_file = os.path.join(state_dir, "uuid.txt")
            with io.open(uuid_file, "w", encoding="UTF-8") as f:
                f.write(self.fsgs.game.variant.uuid + "\n")
        # timestamp_file = os.path.join(state_dir, "timestamp.txt")
        # with open(timestamp_file, "wb") as f:
        #     f.write("\n")

        return state_dir

    def emulator_state_dir(self, emulator):
        assert emulator
        state_dir = os.path.join(self.get_state_dir(), emulator)
        if not os.path.exists(state_dir):
            os.makedirs(state_dir)
        return state_dir

    def get_game_file(self, config_key="cartridge"):
        if self.__game_temp_file is not None:
            return self.__game_temp_file.path

        file_uri = self.config[config_key]
        # FIXME: create new API fsgs.file.path(file_uri), returns path
        # (and creates temp file if necessary)
        input_stream = self.fsgs.file.open(file_uri)

        self.__game_temp_file = self.fsgs.temp_file(file_uri.split("/")[-1])

        with open(self.__game_temp_file.path, "wb") as f:
            f.write(input_stream.read())
        return self.__game_temp_file.path

    def create_temp_dir(self, suffix):
        return self.fsgs.temp_dir(suffix)

    def create_temp_file(self, suffix):
        return self.fsgs.temp_file(suffix)

    def set_env(self, name, value):
        self.__env[name] = value

    def add_arg(self, *args):
        self.__args.extend(args)

    def start_emulator_from_plugin_resource(
            self, provide_name, args=None, env_vars=None):
        resource = self.fsgs.plugins.find_executable(provide_name)
        return self.start_emulator("", args=args, env_vars=env_vars,
                                   executable=resource.path)

    def start_emulator(
            self, emulator, args=None, env_vars=None, executable=None,
            cwd=None):
        if "/" in emulator:
            if not executable:
                executable = self.find_emulator_executable(emulator)
            if not executable:
                emulator = emulator.split("/")[-1]

        if not executable:
            executable = self.find_emulator_executable("fs-" + emulator)
        if not executable:
            executable = self.find_emulator_executable(emulator)

        if not executable:
            raise Exception("could not find executable for " + emulator)
        process_args = [executable]
        process_args.extend(self.__args)
        if args is not None:
            process_args.extend(args)
        print(repr(process_args))

        if "SDL_VIDEODRIVER" in os.environ:
            print("SDL_VIDEODRIVER was present in environment, removing!")
            del os.environ["SDL_VIDEODRIVER"]

        env = os.environ.copy()
        if self.use_fullscreen():
            for key in ["FSGS_GEOMETRY", "FSGS_WINDOW"]:
                if env.get(key, ""):
                    # Already specified externally, so we just use
                    # the existing values...
                    pass
                    print("using existing fullscreen window rect")
                else:
                    x, y, w, h = self.screen_rect()
                    env[key] = "{0},{1},{2},{3}".format(x, y, w, h)
            print("fullscreen rect:", env.get("FSGS_GEOMETRY", ""))

            if self.config.get("fullscreen_mode", "") == "window":
                print("using fullscreen window mode")
                env["FSGS_FULLSCREEN"] = "window"
            elif self.config.get("fullscreen_mode", "") == "fullscreen":
                print("using fullscreen (legacy) mode")
                env["FSGS_FULLSCREEN"] = "fullscreen"
            else:
                print("using fullscreen desktop mode")
                env["FSGS_FULLSCREEN"] = "desktop"
        else:
            print("using window mode (no fullscreen)")
            env["FSGS_FULLSCREEN"] = ""

        env.update(self.__env)
        env["HOME"] = self.home.path

        if env_vars:
            env.update(env_vars)
        print(env)

        args = [process_args]
        kwargs = {}
        if env is not None:
            kwargs["env"] = env
        if cwd is not None:
            kwargs["cwd"] = cwd
        else:
            kwargs["cwd"] = self.cwd.path
        if windows:
            kwargs["close_fds"] = True
        print(" ".join(process_args))
        current_task.set_progress(
            "Starting {emulator}".format(emulator=emulator))
        process = subprocess.Popen(*args, **kwargs)
        return process

    def find_emulator_executable(self, name):
        # if os.path.isdir("../fs-uae/src"):
        #     # running from source directory, we then want to find locally
        #     # compiled binaries if available
        #     path = "../fs-uae/fs-uae"
        #     if windows:
        #         path += ".exe"
        #     if os.path.isfile(path):
        #         return path
        #     raise Exception("Could not find development FS-UAE executable")

        if "/" in name:
            package, name = name.split("/")
        else:
            package = name

        if windows:
            # first we check if the emulator is bundled inside the launcher
            # directory
            exe = os.path.join(
                app.executable_dir(), package, name + ".exe")
            if not os.path.exists(exe):
                # for when the emulators are placed alongside the launcher /
                # game center directory
                exe = os.path.join(
                    app.executable_dir(), "..", package, name + ".exe")
            if not os.path.exists(exe):
                # when the emulators are placed alongside the fs-uae/ directory
                # containing launcher/, for FS-UAE Launcher & FS-UAE Arcade
                exe = os.path.join(
                    app.executable_dir(), "..", "..", package, name + ".exe")
            if not os.path.exists(exe):
                return None
            return exe
        elif macosx:
            exe = os.path.join(
                app.executable_dir(), "..",
                package + ".app", "Contents", "MacOS", name)
            if not os.path.exists(exe):
                exe = os.path.join(
                    app.executable_dir(), "..", "..", "..",
                    package + ".app", "Contents", "MacOS", name)
            if not os.path.exists(exe):
                exe = os.path.join(
                    "/Applications",
                    package + ".app", "Contents", "MacOS", name)
            if not os.path.exists(exe):
                return None
            return exe

        if os.path.exists("/opt/{0}/bin/{1}".format(package, name)):
            return "/opt/{0}/bin/{1}".format(package, name)

        if package == name:
            for dir in os.environ["PATH"].split(":"):
                # dir = unicode_path(dir)
                path = os.path.join(dir, name)
                if os.path.exists(path):
                    return path
        return None

    def get_game_refresh_rate(self):
        """Override in inherited classes to specify custom refresh rate."""
        if self.config["refresh_rate"]:
            return float(self.config["refresh_rate"])
        if self.config.get("video_standard") == "NTSC":
            return 59.94
        if self.config.get("video_standard") == "PAL":
            return 50
        if "USA" in self.fsgs.game.variant.name:
            return 59.94
        if "World" in self.fsgs.game.variant.name:
            return 59.94
        if "Japan" in self.fsgs.game.variant.name:
            return 59.94
        if "Europe" in self.fsgs.game.variant.name:
            return 50

    def configure_vsync(self):
        print("\n" + "-" * 79 + "\n" + "CONFIGURE VSYNC")
        # print("^" * 80)
        allow_vsync = self.use_vsync()
        # if self.get_option('vsync'):
        if allow_vsync:
            # try:
            #     game_refresh = float(self.config["refresh_rate"])
            # except Exception:
            game_refresh = self.get_game_refresh_rate() or 0.0
            # if self.context.game.config.get('system', '') == 'NTSC':
            #     game_refresh = 60.0
            # else:
            #     game_refresh = 50.0
            refresh_rate_tool = RefreshRateTool(
                game_platform=self.fsgs.game.platform.id,
                game_refresh=round(game_refresh))
            refresh_rate_tool.set_best_mode()
            if refresh_rate_tool.allow_vsync():
                print("setting vsync to true")
                self.set_vsync(True)
                return refresh_rate_tool.get_display_refresh()
        else:
            print("vsync is not enabled")
        print("setting vsync to false")
        self.set_vsync(False)
        return False

    def set_vsync(self, vsync):
        print("GameRunner - set vsync to", vsync)
        self.__vsync = vsync

    def abort(self):
        print("GameRunner.abort - WARNING: not implemented")



import fsui as fsui
from .Skin import Skin


class CustomPanel(fsui.Panel):

    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        Skin.set_background_color(self)
        self.layout = fsui.VerticalLayout()



import math
import datetime
from fsui.qt import Qt, QRect, QLabel, QVBoxLayout, QWidget
from fsui.qt import QPainter, QBrush, QPen, QColor
from fs_uae_workspace.shell import QWindow, SimpleApplication


class ClockWindow(QWindow):

    def __init__(self):
        QWindow.__init__(self, None)
        self.setWindowTitle("Clock")
        layout = QVBoxLayout()
        self.label = QLabel("00:00")
        self.widget = ClockWidget()
        # self.label.show()
        # self.label.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
        layout.addWidget(self.widget)
        self.setLayout(layout)

        self.resize(200, 200)

    def __del__(self):
        print("ClockWindow.__del__")


class ClockWidget(QWidget):

    def __init__(self):
        QWidget.__init__(self, None)
        self.setWindowTitle("Clock")

        self.time = None
        self.updateClock()
        self.startTimer(100)

    # noinspection PyPep8Naming
    def updateClock(self):
        now = datetime.datetime.now()
        new_time = (now.hour, now.minute, now.second)
        if new_time != self.time:
            self.time = new_time
            self.repaint()

    def timerEvent(self, event):
        self.updateClock()
        import gc
        print(gc.garbage)

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setRenderHints(QPainter.Antialiasing)
        pen = QPen(QColor(0x80, 0x80, 0x80))
        pen.setWidth(2)
        painter.setPen(pen)
        painter.setBrush(QBrush(Qt.white))
        x, y, w, h = 10, 10, self.width() - 20, self.height() - 20
        rect = QRect(x, y, w, h)
        # painter.drawChord(rect, 0, 16 * 360)
        painter.drawEllipse(rect)

        cx = x + w / 2
        cy = y + h / 2
        r = w / 2 - 10

        pen.setWidth(4)
        painter.setPen(pen)

        for i in range(12):
            px = cx + r * math.cos(2 * math.pi * i / 12)
            py = cy + r * math.sin(2 * math.pi * i / 12)
            painter.drawPoint(px, py)

        hours, minutes, seconds = self.time
        minutes += seconds / 60.0
        hours += minutes / 60.0

        f = 2.0 * math.pi
        f2 = - math.pi / 2.0

        r = w / 2 * 0.6
        # pen = QPen(QColor(0xff, 0x00, 0x00))
        pen.setWidth(2)
        painter.setPen(pen)
        painter.drawLine(cx, cy,
                         cx + r * math.cos(f * hours / 12 + f2),
                         cy + r * math.sin(f * hours / 12 + f2))

        r = w / 2 * 0.8
        # pen = QPen(QColor(0xff, 0x00, 0x00))
        pen.setWidth(2)
        painter.setPen(pen)
        painter.drawLine(cx, cy,
                         cx + r * math.cos(f * minutes / 60 + f2),
                         cy + r * math.sin(f * minutes / 60 + f2))

        pen = QPen(QColor(0xff, 0x00, 0x00))
        # pen.setWidth(2)
        painter.setPen(pen)
        painter.drawLine(cx, cy,
                         cx + r * math.cos(f * seconds / 60 + f2),
                         cy + r * math.sin(f * seconds / 60 + f2))

application = SimpleApplication(ClockWindow)
